# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Optimization Rule - Correlated Filters

Type: Heuristic
Goal: Reduce Rows

When fields are joined on, we can infer ranges of values based on statistics
or filters. This can be used to reduce the number of rows that need to be read
and processed.
"""

from orso.tools import random_string

from opteryx.managers.expression import NodeType
from opteryx.models import Node
from opteryx.planner import build_literal_node
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext
from .optimization_strategy import get_nodes_of_type_from_logical_plan


def _write_filters(left_column, right_column):
    new_filters = []
    if left_column.schema_column.highest_value is not None:
        a_side = right_column
        b_side = build_literal_node(left_column.schema_column.highest_value)
        new_filter = Node(
            LogicalPlanStepType.Filter,
            condition=Node(NodeType.COMPARISON_OPERATOR, value="LtEq", left=a_side, right=b_side),
            columns=[right_column],
            relations={right_column.source},
            all_relations={right_column.source},
        )
        new_filters.append(new_filter)

        a_side = right_column
        b_side = build_literal_node(left_column.schema_column.lowest_value)
        new_filter = Node(
            LogicalPlanStepType.Filter,
            condition=Node(NodeType.COMPARISON_OPERATOR, value="GtEq", left=a_side, right=b_side),
            columns=[right_column],
            relations={right_column.source},
            all_relations={right_column.source},
        )
        new_filters.append(new_filter)
    return new_filters


class CorrelatedFiltersStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if (
            node.node_type == LogicalPlanStepType.Join
            and node.type == "inner"
            and len(node.all_relations) == 2
        ):
            left_column = node.on.left
            right_column = node.on.right
            new_filters = []

            # Empty connectors are FUNCTION datasets, we could push filters down and create
            # statistics for them, but there are other issues this creates
            if (
                left_column.node_type == NodeType.IDENTIFIER
                and right_column.node_type == NodeType.IDENTIFIER
                and left_column.source_connector != set()
            ):
                new_filters = _write_filters(left_column, right_column)
            if (
                left_column.node_type == NodeType.IDENTIFIER
                and right_column.node_type == NodeType.IDENTIFIER
                and right_column.source_connector != set()
            ):
                new_filters.extend(_write_filters(right_column, left_column))
            for new_filter in new_filters:
                context.optimized_plan.insert_node_before(
                    random_string(), new_filter, context.node_id
                )
                self.statistics.optimization_inner_join_correlated_filter += 1

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan

    def should_i_run(self, plan):
        # only run if there are LIMIT clauses in the plan
        candidates = get_nodes_of_type_from_logical_plan(plan, (LogicalPlanStepType.Join,))
        return len(candidates) > 0



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

import numpy
from orso.types import OrsoTypes

from opteryx.exceptions import InvalidFunctionParameterError
from opteryx.exceptions import SqlError
from opteryx.utils import dates


def generate_series(*args):
    arg_len = len(args)
    arg_vals = [i.value for i in args]
    first_arg_type = args[0].type

    # if the parameters are numbers, generate series is an alias for range
    if first_arg_type.is_numeric():
        if arg_len not in (1, 2, 3):  # pragma: no cover
            raise SqlError(
                "GENERATE_SERIES for numbers takes 1 (stop), 2 (start, stop) or 3 (start, stop, interval) parameters."
            )
        return numeric_range(*arg_vals)

    # if the params are timestamps, we create time intervals
    if args[0].type in (OrsoTypes.DATE, OrsoTypes.TIMESTAMP):
        if arg_len != 3:  # pragma: no cover
            raise SqlError("generate_series for dates needs start, end, and interval parameters")
        return dates.date_range(*arg_vals)

    raise InvalidFunctionParameterError(
        "Unsupported value for GENERATE_SERIES, must be date or numeric series."
    )


def numeric_range(*args) -> numpy.ndarray:
    """
    Generate a numeric range of vales

    Args:
        [start, ]stop, [step, ]: Arguments as in numpy.arange.

    Returns:
        numpy.ndarray: Array of evenly spaced values.

    Raises:
        ValueError: If the number of arguments is not 1, 2, or 3.

    Examples:
        generate_range(5)
        generate_range(1, 5)
        generate_range(1, 5, 0.5)
    """
    # Define defaults
    start, step, dtype = numpy.int64(1), numpy.int64(1), numpy.float64

    # Process arguments
    if len(args) == 1:
        stop = args[0]
    elif len(args) == 2:
        start, stop = args
    elif len(args) == 3:
        start, stop, step = args
    else:  # pragma: no cover
        raise ValueError("Invalid number of arguments. Expected 1, 2, or 3: start, stop [, step].")

    # Determine dtype
    if all(numpy.issubdtype(arg, numpy.integer) for arg in [start, stop, step]):
        dtype = numpy.int64  # type: ignore

    # Compute range
    num_range = numpy.arange(start, stop + step, step, dtype=dtype)

    # Check last value, remove if it doesn't fall on a step boundary or is over the stop value
    if not numpy.isclose(num_range[-1], stop, atol=step / 2) or num_range[-1] > stop:  # type: ignore
        num_range = num_range[:-1]

    return num_range



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
The file connector provides the reader for when a file name is provided as the
dataset name in a query.
"""

import glob
import mmap
import os
from typing import Dict
from typing import List
from typing import Optional

import pyarrow
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import LimitPushable
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.connectors.capabilities import Statistics
from opteryx.exceptions import DatasetNotFoundError
from opteryx.utils import is_windows
from opteryx.utils.file_decoders import get_decoder

IS_WINDOWS = is_windows()
# Define os.O_BINARY for non-Windows platforms if it's not already defined
if not hasattr(os, "O_BINARY"):
    os.O_BINARY = 0  # Value has no effect on non-Windows platforms
if not hasattr(os, "O_DIRECT"):
    os.O_DIRECT = 0  # Value has no effect on non-Windows platforms

mmap_config = {}
if not IS_WINDOWS:
    mmap_config["flags"] = mmap.MAP_PRIVATE
    mmap_config["prot"] = mmap.PROT_READ
else:
    mmap_config["access"] = mmap.ACCESS_READ


def read_blob(
    *, blob_name: str, decoder, just_schema=False, statistics=None, projection=None, selection=None
):
    """
    Read a blob (binary large object) from disk using memory-mapped file access.

    This method uses low-level file reading with memory-mapped files to
    improve performance. It reads the entire file into memory and then
    decodes it using the provided decoder function.

    Parameters:
        blob_name (str):
            The name of the blob file to read.
        decoder (callable):
            A function to decode the memory-mapped file content.
        just_schema (bool, optional):
            If True, only the schema of the data is returned. Defaults to False.
        projection (list, optional):
            A list of fields to project. Defaults to None.
        selection (dict, optional):
            A dictionary of selection criteria. Defaults to None.
        **kwargs:
            Additional keyword arguments.

    Returns:
        The decoded blob content.

    Raises:
        FileNotFoundError:
            If the blob file does not exist.
        OSError:
            If an I/O error occurs while reading the file.
    """
    try:
        file_descriptor = os.open(blob_name, os.O_RDONLY | os.O_BINARY)
        if hasattr(os, "posix_fadvise"):
            os.posix_fadvise(file_descriptor, 0, 0, os.POSIX_FADV_WILLNEED)
        size = os.fstat(file_descriptor).st_size
        _map = mmap.mmap(file_descriptor, length=size, **mmap_config)
        result = decoder(
            _map,
            just_schema=just_schema,
            projection=projection,
            selection=selection,
            use_threads=True,
        )
        statistics.bytes_read += size

        return result
    finally:
        os.close(file_descriptor)


class FileConnector(BaseConnector, PredicatePushable, Statistics, LimitPushable):
    """
    Connector for reading datasets from a file.
    """

    __mode__ = "Blob"
    __type__ = "FILE"
    _byte_array: Optional[bytes] = None  # Instance attribute to store file bytes

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        "NotEq": True,
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
    }

    PUSHABLE_TYPES = {
        OrsoTypes.BLOB,
        OrsoTypes.BOOLEAN,
        OrsoTypes.DOUBLE,
        OrsoTypes.INTEGER,
        OrsoTypes.VARCHAR,
        OrsoTypes.TIMESTAMP,
        OrsoTypes.DATE,
    }

    @property
    def interal_only(self):
        return True

    def __init__(self, *args, **kwargs):
        BaseConnector.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)
        Statistics.__init__(self, **kwargs)
        LimitPushable.__init__(self, **kwargs)

        if ".." in self.dataset or self.dataset[0] in ("\\", "/", "~"):
            # Don't find any datasets which look like path traversal
            raise DatasetNotFoundError(dataset=self.dataset)

        # Check if dataset contains wildcards
        self.has_wildcards = any(char in self.dataset for char in ["*", "?", "["])

        if self.has_wildcards:
            # Expand wildcards to get list of files
            self.files = self._expand_wildcards(self.dataset)
            if not self.files:
                raise DatasetNotFoundError(dataset=self.dataset)
            # Use the first file to determine the decoder
            self.decoder = get_decoder(self.files[0])
        else:
            self.files = [self.dataset]
            self.decoder = get_decoder(self.dataset)

    def _expand_wildcards(self, pattern: str) -> List[str]:
        """
        Expand wildcard patterns in file paths while preventing path traversal.

        Supports wildcards:
        - * matches any number of characters
        - ? matches a single character
        - [range] matches a range of characters (e.g., [0-9], [a-z])

        Args:
            pattern: File path pattern with wildcards

        Returns:
            List of matching file paths
        """
        # Additional path traversal check after expansion
        if ".." in pattern:
            raise DatasetNotFoundError(dataset=pattern)

        # Use glob to expand the pattern
        matched_files = glob.glob(pattern, recursive=False)

        # Filter out any results that might have path traversal
        # This is an extra safety check
        safe_files = []
        for file_path in matched_files:
            if ".." not in file_path and os.path.isfile(file_path):
                safe_files.append(file_path)

        return sorted(safe_files)

    def read_dataset(
        self, columns: list = None, predicates: list = None, limit: int = None, **kwargs
    ) -> pyarrow.Table:
        rows_read = 0

        # Iterate over all matched files
        for file_path in self.files:
            morsel = read_blob(
                blob_name=file_path,
                decoder=self.decoder,
                statistics=self.statistics,
                projection=columns,
                selection=predicates,
            )[3]

            if limit is not None:
                remaining = limit - rows_read
                if remaining <= 0:
                    break
                if morsel.num_rows > remaining:
                    morsel = morsel.slice(offset=0, length=remaining)
                rows_read += morsel.num_rows

            yield morsel

    def get_dataset_schema(self) -> RelationSchema:
        """
        Retrieves the schema from the dataset file.

        Returns:
            The schema of the dataset.
        """
        import mmap

        if self.schema is not None:
            return self.schema

        # Use the first file to get the schema
        first_file = self.files[0]

        try:
            file_descriptor = os.open(first_file, os.O_RDONLY | os.O_BINARY)
            size = os.path.getsize(first_file)
            _map = mmap.mmap(file_descriptor, size, access=mmap.ACCESS_READ)
            self.schema = self.decoder(_map, just_schema=True)
            self.relation_statistics = self.decoder(_map, just_statistics=True)
            return self.schema
        finally:
            os.close(file_descriptor)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Optimization Rule - Join Rewriter

Type: Heuristic
Goal: Faster Joins

Rewrite joins to more efficient versions based on heuristics.

We collect references to LEFT JOINs and the right relation as we traverse the plan.
"""

from opteryx.managers.expression import NodeType
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext
from .optimization_strategy import get_nodes_of_type_from_logical_plan

LEFT_REWRITABLE_CONDITIONS = ("Eq", "Gt", "Lt", "Gte", "Lte")


class JoinRewriteStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Filter:
            context.collected_predicates.append(node)

        if node.node_type == LogicalPlanStepType.Join and node.type == "left outer":
            context.collected_joins.append(node)

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        if context.collected_joins and context.collected_predicates:
            for join in context.collected_joins:
                left_side = join.left_relation_names
                can_rewrite = None
                for predicate in context.collected_predicates:
                    if predicate.condition.value in LEFT_REWRITABLE_CONDITIONS and set(
                        left_side
                    ).intersection(predicate.all_relations):
                        if (
                            predicate.condition.left.node_type == NodeType.IDENTIFIER
                            and predicate.condition.right.node_type == NodeType.IDENTIFIER
                        ):
                            import warnings

                            warnings.warn("LEFT OUTER JOIN -> INNER JOIN rewriter not implemented")
                    if predicate.condition.value == "IsNull" and set(left_side).intersection(
                        predicate.all_relations
                    ):
                        import warnings

                        warnings.warn("LEFT OUTER JOIN -> SEMI JOIN rewriter not implemented")
                    if predicate.condition.value == "IsNotNull" and set(left_side).intersection(
                        predicate.all_relations
                    ):
                        import warnings

                        warnings.warn("LEFT OUTER JOIN -> ANTI JOIN rewriter not implemented")
        return plan

    def should_i_run(self, plan):
        # only run if there are LIMIT clauses in the plan
        candidates = get_nodes_of_type_from_logical_plan(plan, (LogicalPlanStepType.Join,))
        return len(candidates) > 0



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Optimization Rule - Empty Table Elimination

Type: Structural
Goal: Replace FILTER(FALSE) + subtree with empty data source

When predicate compaction or other optimizations detect contradictory
conditions (e.g., column = 'A' AND column = 'B'), they produce FILTER(FALSE).
Rather than executing the entire subtree and filtering out all rows, we
detect these impossible conditions and replace the FILTER node and its
subtree with an empty FunctionDataset node.

This is purely a structural optimization that reduces execution overhead.
"""

from opteryx.managers.expression import NodeType
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext


class EmptyTableStrategy(OptimizationStrategy):  # pragma: no cover
    """
    Replace FILTER(FALSE) with empty data source.

    This strategy detects impossible filter conditions (contradictions)
    and replaces the FILTER node and its entire subtree with a FunctionDataset
    that returns an empty table with the correct schema.

    During visit(), we collect FALSE filters. In complete(), we safely rewrite
    them by replacing with empty data sources.
    """

    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        """
        During traversal, collect FILTER nodes that have FALSE conditions.
        Don't modify the graph here - just collect for later processing.
        """
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Filter:
            if self._is_false_condition(node.condition):
                # Store the node ID and the node itself for later rewriting
                context.false_filters.append((context.node_id, node))

        return context

    def should_i_run(self, plan: LogicalPlan) -> bool:
        """Only run if there are FILTER clauses with FALSE conditions in the plan."""
        for _, node in plan.nodes(True):
            if node.node_type == LogicalPlanStepType.Filter:
                if self._is_false_condition(node.condition):
                    return True
        return False

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        """
        Post-process the optimized plan to replace FILTER(FALSE) with empty data source.

        We collected FALSE filters during visit(). Now safely replace them by:
        1. Finding all descendants of the filter node (NOT including the filter itself)
        2. Removing descendants one by one
        3. Replacing the filter node with an empty FunctionDataset
        """
        optimized_plan = context.optimized_plan

        # Process each FALSE filter we collected
        for filter_nid, _ in context.false_filters:
            # Collect all descendants starting from outgoing edges (NOT the filter itself)
            descendants = self._collect_descendants(filter_nid, optimized_plan)

            # Get all properties from the immediate child (Scan node) before removing it
            # This preserves the schema, columns, connector, etc.
            child_properties = {}
            if descendants:
                # The first descendant is the direct child (typically a Scan node)
                first_child_nid = descendants[0]
                first_child_node = optimized_plan[first_child_nid]
                # Copy all properties from the child node
                child_properties = dict(first_child_node.properties)

            # Remove descendants one by one, from leaf up (reverse order)
            for desc_nid in reversed(descendants):
                optimized_plan.remove_node(desc_nid, heal=True)

            # Create a Scan node with marker for empty result
            # The physical planner will see this marker and use NullReaderNode
            # Start with child properties and override connector with our marker
            empty_props = child_properties.copy()
            empty_props["node_type"] = LogicalPlanStepType.Scan
            empty_props["connector"] = (
                "__null__"  # Marker for physical planner to use NullReaderNode
            )

            empty_node = LogicalPlanNode(**empty_props)

            # Replace the filter node with the empty node (node must still exist in graph)
            optimized_plan[filter_nid] = empty_node

            self.statistics.optimization_empty_table += 1

        return optimized_plan

    def _collect_descendants(self, nid: str, plan: LogicalPlan) -> list:
        """
        Collect all descendant nodes starting from the given node.
        Returns a list of node IDs to remove.
        """
        descendants = []
        stack = list(
            plan.outgoing_edges(nid)
        )  # Start with direct children - returns (source, target, relationship)
        visited = {nid}

        while stack:
            _, current_nid, _ = stack.pop()  # Unpack (source, target, relationship)

            if current_nid not in visited:
                visited.add(current_nid)
                descendants.append(current_nid)

                # Add this node's children to the stack
                stack.extend(plan.outgoing_edges(current_nid))

        return descendants

    def _is_false_condition(self, condition) -> bool:
        """Check if a condition node is FALSE literal."""
        if condition is None:
            return False
        return condition.node_type == NodeType.LITERAL and condition.value is False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Utilities for detecting and handling Python's free-threading mode.
"""

import sys


def is_free_threading_available() -> bool:
    """
    Detect if Python is running with free-threading (GIL disabled).

    Python 3.13+ supports free-threading mode where the GIL can be disabled.
    This function detects if that mode is active using multiple detection methods.

    Returns:
        bool: True if free-threading is available and enabled, False otherwise.
    """
    # Method 1: Check sys._is_gil_disabled() (Python 3.13+)
    if hasattr(sys, "_is_gil_disabled"):
        if callable(sys._is_gil_disabled):
            return sys._is_gil_disabled()
        # In some builds it might be a bool attribute
        return bool(sys._is_gil_disabled)

    # Method 2: Check sysconfig for Py_GIL_DISABLED (most reliable for 3.13)
    try:
        import sysconfig

        gil_disabled = sysconfig.get_config_var("Py_GIL_DISABLED")
        if gil_disabled is not None:
            return gil_disabled == 1
    except (ImportError, Exception):
        pass

    # Method 3: Check sys.flags.gil (newer Python 3.13 builds)
    if hasattr(sys, "flags") and hasattr(sys.flags, "gil"):
        return sys.flags.gil == 0

    # Fallback: no free-threading support detected
    return False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
derived
---------

This is used as the source relation for derived values.
"""

from orso.schema import RelationSchema

from opteryx.models import RelationStatistics

__all__ = ("schema",)


def schema():
    return RelationSchema(name="$derived", columns=[])


def statistics() -> RelationStatistics:
    return RelationStatistics()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Simple Grouping Node

This is a SQL Query Execution Plan Node.

This is the grouping node, this specialized version only performs aggregations that result in,
and are collected as, a single value.
"""

import time

import numpy
import pyarrow
from orso.types import OrsoTypes

from opteryx import EOS
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate_and_append
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import QueryProperties
from opteryx.operators.aggregate_node import AGGREGATORS
from opteryx.operators.aggregate_node import build_aggregations
from opteryx.operators.aggregate_node import extract_evaluations
from opteryx.operators.aggregate_node import project

from . import BasePlanNode

CHUNK_SIZE = 65536


def build_finalizer_aggregations(aggregators):
    column_map = {}
    aggs = []

    if not isinstance(aggregators, list):
        aggregators = [aggregators]

    for root in aggregators:
        for aggregator in get_all_nodes_of_type(root, select_nodes=(NodeType.AGGREGATOR,)):
            count_options = None

            field_name = aggregator.schema_column.identity
            if aggregator.value == "COUNT":
                function = AGGREGATORS["SUM"]
            else:
                function = AGGREGATORS[aggregator.value]
            # if the array agg is distinct, base off that function instead
            aggs.append((field_name, function, count_options))
            column_map[aggregator.schema_column.identity] = f"{field_name}_{function}".replace(
                "_hash_", "_"
            )

    return column_map, aggs


class SimpleAggregateAndGroupNode(BasePlanNode):
    SIMPLE_AGGREGATES = {"SUM", "MIN", "MAX", "COUNT", "AVG", "COUNT_DISTINCT"}

    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)
        self.groups = list(parameters["groups"])
        self.aggregates = list(parameters["aggregates"])
        projection = list(parameters["projection"])

        # Replace offset based GROUP BYs with their column
        self.groups = [
            (
                group
                if not (group.node_type == NodeType.LITERAL and group.type == OrsoTypes.INTEGER)
                else projection[group.value - 1]
            )
            for group in self.groups
        ]

        # get all the columns anywhere in the groups or aggregates
        all_identifiers = [
            node.schema_column.identity
            for node in get_all_nodes_of_type(
                self.groups + self.aggregates, select_nodes=(NodeType.IDENTIFIER,)
            )
        ]
        self.all_identifiers = list(dict.fromkeys(all_identifiers))

        # Get any functions we need to execute before aggregating
        self.evaluatable_nodes = extract_evaluations(self.aggregates)

        # get the aggregated groupings and functions
        self.group_by_columns = list({node.schema_column.identity for node in self.groups})
        self.column_map, self.aggregate_functions = build_aggregations(self.aggregates)

        self.finalizer_map, self.finalizer_aggregations = build_finalizer_aggregations(
            self.aggregates
        )

        self.buffer = []

    @property
    def config(self):  # pragma: no cover
        from opteryx.managers.expression import format_expression

        return f"AGGREGATE ({', '.join(format_expression(col) for col in self.aggregates)}) GROUP BY ({', '.join(format_expression(col) for col in self.groups)})"

    @property
    def name(self):  # pragma: no cover
        return "Group By Simple"

    def execute(self, morsel: pyarrow.Table, **kwargs):
        internal_names = list(self.column_map.values()) + self.group_by_columns
        column_names = list(self.column_map.keys()) + self.group_by_columns

        if morsel == EOS:
            start = time.monotonic_ns()

            internal_names = list(self.finalizer_map.values()) + self.group_by_columns
            column_names = list(self.finalizer_map.keys()) + self.group_by_columns

            groups = pyarrow.concat_tables(self.buffer, promote_options="permissive")
            self.buffer.clear()
            groups = groups.group_by(self.group_by_columns)
            groups = groups.aggregate(self.finalizer_aggregations)
            groups = groups.select(internal_names)
            groups = groups.rename_columns(column_names)

            self.statistics.time_groupby_finalize += time.monotonic_ns() - start

            num_rows = groups.num_rows
            for start in range(0, num_rows, CHUNK_SIZE):
                yield groups.slice(start, min(CHUNK_SIZE, num_rows - start))

            yield EOS
            return

        morsel = project(morsel, self.all_identifiers)

        # Allow grouping by functions by evaluating them first
        if self.evaluatable_nodes:
            morsel = evaluate_and_append(self.evaluatable_nodes, morsel)

        morsel = evaluate_and_append(self.groups, morsel)

        # Add a "*" column, this is an int because when a bool it miscounts
        if "*" not in morsel.column_names:
            morsel = morsel.append_column(
                "*", [numpy.full(shape=morsel.num_rows, fill_value=1, dtype=numpy.int8)]
            )

        # use pyarrow to do phase 1 of the group by
        st = time.monotonic_ns()
        groups = morsel.group_by(self.group_by_columns)
        groups = groups.aggregate(self.aggregate_functions)
        groups = groups.select(internal_names)
        groups = groups.rename_columns(column_names)
        self.statistics.time_pregrouping += time.monotonic_ns() - st

        self.buffer.append(groups)

        yield None



from opteryx.config import features
from opteryx.exceptions import InvalidInternalStateError
from opteryx.utils.threading import is_free_threading_available

from .parallel_engine import execute as parallel_execute
from .serial_engine import execute as serial_execute

ENABLE_FREE_THREADING = features.enable_free_threading


def execute(plan, statistics):
    # Validate query plan to ensure it's acyclic
    if not plan.is_acyclic():
        raise InvalidInternalStateError("Query plan is cyclic, cannot execute.")

    # Label the join legs to ensure left/right ordering
    plan.label_join_legs()

    # Use parallel engine if free-threading is available, otherwise use serial
    if ENABLE_FREE_THREADING and is_free_threading_available():
        # DEBUG: print("\033[38;2;255;184;108mUsing parallel execution engine.\033[0m")
        yield from parallel_execute(plan, statistics=statistics)
    else:
        yield from serial_execute(plan, statistics=statistics)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Optimization Rule - Predicate Ordering

Type: Cost
Goal: Faster Execution

We combine adjacent predicates into chains of ANDed conditions in a single
filtering step. We use cost estimates based on the data type to determine
a good order to execute the filters.

NOTE: This currently doesn't account for a few very important things which
means this may create an order that is slower than the order that would have
been run if this strategy didn't run.

- The selectivity of the filter
- How to handle complex sub conditions or ORed conditions
"""

from orso.schema import ConstantColumn
from orso.tools import random_string

# pragma: no cover
from orso.types import OrsoTypes

from opteryx.managers.expression import NodeType
from opteryx.models import Node
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext
from .optimization_strategy import get_nodes_of_type_from_logical_plan

# Approximate of the time in seconds (3dp) to compare 1 million records
# None indicates no comparison is possible
# These are the core comparisons, Eq, NotEq, Gt, GtEq, Lt, LtEq
BASIC_COMPARISON_COSTS = {
    OrsoTypes.ARRAY: 10.00,  # expensive
    OrsoTypes.BLOB: 0.06,  # varies based on length, this is 50 bytes
    OrsoTypes.JSONB: 10.00,  # JSONB (treat as expensive)
    OrsoTypes.BOOLEAN: 0.004,
    OrsoTypes.DATE: 0.01,
    OrsoTypes.DECIMAL: 2.35,
    OrsoTypes.DOUBLE: 0.003,
    OrsoTypes.INTEGER: 0.002,
    OrsoTypes.INTERVAL: 10.00,  # expensive
    OrsoTypes.STRUCT: 10.00,  # expensive
    OrsoTypes.TIMESTAMP: 0.009,
    OrsoTypes.TIME: 10.00,  # expensive
    OrsoTypes.VARCHAR: 0.3,  # varies based on length, this is 50 chars
    OrsoTypes.NULL: 10.00,  # for completeness
    OrsoTypes._MISSING_TYPE: 10.00,  # for completeness
    0: 10.00,  # for completeness
}


def rewrite_anded_any_eq_to_contains_all(predicate, statistics):
    """
    Rewrite multiple AND'ed ANYOPEQ conditions on the same column into a single ArrayContainsAll (@>>) condition.

    Example:
      'a' = ANY(z) AND 'b' = ANY(z) AND 'c' = ANY(z)
      -->  z @>> ('a','b','c')     # BinaryOperator::Custom("ArrayContainsAll")

    Notes:
      - We only match: LITERAL = ANY(IDENTIFIER)
      - We group by the SAME column identity
      - Remaining AND nodes are neutralized to TRUE (since X AND TRUE == X)
    """
    anyeq_by_col = {}

    def collect_any_eq_and(node, grouped):
        # Only collect beneath ANDs (like your OR rewrite only walks ORs)
        if node.node_type == NodeType.DNF:
            for param in node.parameters:
                if param.node_type == NodeType.COMPARISON_OPERATOR and param.value == "AnyOpEq":
                    # literal = ANY(identifier)
                    if (
                        param.left.node_type == NodeType.LITERAL
                        and param.right.node_type == NodeType.IDENTIFIER
                    ):
                        col_id = param.right.schema_column.identity
                        if col_id not in grouped:
                            grouped[col_id] = {
                                "values": [],
                                "nodes": [],
                                "column_node": param.right,
                            }
                        grouped[col_id]["values"].append(param.left.value)
                        grouped[col_id]["nodes"].append(param)

    collect_any_eq_and(predicate, anyeq_by_col)

    for data in anyeq_by_col.values():
        # Only worth rewriting if we have 2+ literals against the same array column
        if len(data["values"]) > 1:
            # optional: new counter; rename if you already have a metric for this
            if hasattr(statistics, "optimization_predicate_rewriter_anyeq_to_contains_all"):
                statistics.optimization_predicate_rewriter_anyeq_to_contains_all += 1

            # Reuse the first matched node as the replacement site
            new_node = data["nodes"][0]

            # Build right-hand side as an ARRAY constant of unique values
            # (use a set to dedupe; order doesn't matter)
            values_set = set(data["values"])
            new_node.left.value = values_set
            new_node.left.element_type = new_node.left.type
            new_node.left.type = OrsoTypes.ARRAY
            new_node.left.schema_column = ConstantColumn(
                name=new_node.left.name,
                type=OrsoTypes.ARRAY,
                element_type=new_node.left.element_type,
                value=new_node.left.value,
            )

            # Turn node into: column @>> ARRAY[...]
            new_node.value = "ArrayContainsAll"  # your @>> operator
            new_node.node_type = NodeType.COMPARISON_OPERATOR
            new_node.right = data["column_node"]

            # Swap so LHS is the column (array), RHS is the values array
            new_node.left, new_node.right = new_node.right, new_node.left

            # Neutralize the remaining AND'ed ANYOPEQ nodes to TRUE
            for node in data["nodes"][1:]:
                node.node_type = NodeType.LITERAL
                node.type = OrsoTypes.BOOLEAN
                node.value = True

    return predicate


def order_predicates(predicates: list, statistics) -> list:
    """
    This is a fairly naive cost-based predicate ordering routine.

    We have the time it takes for each comparison based on the datatype
    of the values - we order the 'cheapest' (fastest) comparisons first
    """
    cost_estimates = [
        (
            10.00
            if c.condition.left is None
            else BASIC_COMPARISON_COSTS[c.condition.left.schema_column.type],
            i,
        )
        for i, c in enumerate(predicates)
    ]
    # Lowest execution times to the front
    cost_estimates.sort(key=lambda x: x[0], reverse=False)
    ordered_predicates = [predicates[c[1]] for c in cost_estimates]
    # Did this optimization change the ordering?
    for i, e in enumerate(cost_estimates):
        if i > 0 and cost_estimates[i - 1][1] > e[1]:
            statistics.optimization_cost_based_predicate_ordering += 1
            break
    return ordered_predicates


class PredicateOrderingStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Filter:
            node.nid = context.node_id
            context.collected_predicates.append(node)
            return context

        if node.node_type != LogicalPlanStepType.Filter and context.collected_predicates:
            if len(context.collected_predicates) == 1:
                context.collected_predicates = []
                return context

            new_node = LogicalPlanNode(LogicalPlanStepType.Filter)
            new_node.condition = Node(node_type=NodeType.DNF)
            context.collected_predicates = order_predicates(
                context.collected_predicates, self.statistics
            )
            new_node.condition.parameters = [c.condition for c in context.collected_predicates]
            new_node.columns = []
            new_node.relations = set()
            new_node.all_relations = set()

            for predicate in context.collected_predicates:
                new_node.columns.extend(predicate.columns)
                new_node.relations.update(predicate.relations)
                new_node.all_relations.update(predicate.all_relations)
                self.statistics.optimization_flatten_filters += 1
                context.optimized_plan.remove_node(predicate.nid, heal=True)

            new_node.condition = rewrite_anded_any_eq_to_contains_all(
                new_node.condition, self.statistics
            )

            context.optimized_plan.insert_node_after(random_string(), new_node, context.node_id)
            context.collected_predicates.clear()

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan

    def should_i_run(self, plan):
        # only run if there are LIMIT clauses in the plan
        candidates = get_nodes_of_type_from_logical_plan(plan, (LogicalPlanStepType.Filter,))
        return len(candidates) > 0



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Distinct Node

This is a SQL Query Execution Plan Node.

This Node eliminates duplicate records.
"""

from pyarrow import Table

from opteryx import EOS
from opteryx.models import QueryProperties

from . import BasePlanNode


class DistinctNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        from opteryx.third_party.abseil.containers import FlatHashSet as HashSet

        BasePlanNode.__init__(self, properties=properties, **parameters)
        self._distinct_on = parameters.get("on")
        if self._distinct_on:
            self._distinct_on = [col.schema_column.identity for col in self._distinct_on]
        self.hash_set = None

    @property
    def config(self):  # pragma: no cover
        return ""

    @property
    def name(self):  # pragma: no cover
        return "Distinction"

    def execute(self, morsel: Table, **kwargs) -> Table:
        from opteryx.compiled.table_ops.distinct import distinct

        # We create a HashSet outside the distinct call, this allows us to pass
        # the hash to each run of the distinct which means we don't need to concat
        # all of the tables together to return a result.
        #
        # Being able to run morsel-by-morsel means if we have a LIMIT clause, we can
        # limit processing

        if morsel == EOS:
            yield EOS
            return

        unique_indexes, self.hash_set = distinct(
            morsel, columns=self._distinct_on, seen_hashes=self.hash_set
        )

        if len(unique_indexes) > 0:
            distinct_table = morsel.take(unique_indexes)
            yield distinct_table
        else:
            distinct_table = morsel.slice(0, 0)
            yield distinct_table



# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
This module provides the execution engine for processing physical plans with an event loop.
"""

import concurrent.futures
import logging
from typing import Any
from typing import Callable
from typing import Generator
from typing import Optional
from typing import Tuple

import pyarrow

from opteryx import EOS
from opteryx.constants import ResultType
from opteryx.exceptions import InvalidInternalStateError
from opteryx.models import PhysicalPlan
from opteryx.models import QueryStatistics
from opteryx.utils.threading import is_free_threading_available

from .serial_engine import explain

# Configure logging for debugging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Detect if free-threading is available
FREE_THREADING_AVAILABLE = is_free_threading_available()


class ExecutionContext:
    """Encapsulates execution state, thread pool, and async task tracking."""

    def __init__(self, num_workers: int = 2):
        self.use_threading = FREE_THREADING_AVAILABLE
        if self.use_threading:
            self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=num_workers)
            self.futures = {}  # Track active futures {future: (node_id, join_leg)}
        else:
            self.thread_pool = None
            self.futures = {}

    def submit_task(
        self, node: Callable, morsel: Optional[pyarrow.Table], join_leg: Optional[str], nid: str
    ) -> None:
        """Submit a task for execution in the thread pool.

        Args:
            node: Callable node to execute
            morsel: Data batch to process
            join_leg: Join identifier if this is part of a join
            nid: Unique node identifier in the plan
        """
        if self.use_threading:
            future = self.thread_pool.submit(node, morsel, join_leg)
            self.futures[future] = (nid, join_leg)
        else:
            # Without free-threading, execute synchronously
            # Store the result with a dummy future object
            class DummyFuture:
                def __init__(self, result):
                    self._result = result

                def result(self):
                    return self._result

            result = node(morsel, join_leg)
            dummy_future = DummyFuture(result)
            self.futures[dummy_future] = (nid, join_leg)

    def process_async_tasks(self, plan):
        """Process completed async tasks and dispatch results."""
        if not self.futures:
            return

        if self.use_threading:
            # Wait for any completed futures without blocking
            done, not_done = concurrent.futures.wait(
                self.futures.keys(), timeout=0, return_when=concurrent.futures.FIRST_COMPLETED
            )

            # If work is still in progress but nothing is complete, indicate that
            if not done and not_done:
                yield None
                return
        else:
            # Without threading, all tasks are already complete (executed synchronously)
            done = list(self.futures.keys())

        # Process all completed futures
        for future in done:
            nid, join_leg = self.futures.pop(future)
            try:
                results = future.result()
                if results is not None:
                    for result in results:
                        if result is not None:
                            for _, child, leg in plan.outgoing_edges(nid):
                                # Only process stateful nodes immediately
                                node = plan[child]
                                if node.is_stateless:
                                    # Queue up stateless nodes
                                    self.submit_task(node, result, leg, child)
                                else:
                                    yield from process_node(plan, child, result, leg, self)

            except Exception as e:
                logger.error(f"Error processing async task for node {nid}: {e}")

    def shutdown(self):
        """Shut down the execution context."""
        if self.use_threading and self.thread_pool:
            self.thread_pool.shutdown(wait=True)


def execute(
    plan: PhysicalPlan, head_node: str = None, statistics: QueryStatistics = None
) -> Tuple[Generator[pyarrow.Table, Any, Any], ResultType]:
    """Main execution entry point."""
    from opteryx.operators import ExplainNode
    from opteryx.operators import SetVariableNode
    from opteryx.operators import ShowCreateNode
    from opteryx.operators import ShowValueNode

    head_nodes = list(set(plan.get_exit_points()))

    if len(head_nodes) != 1:
        raise InvalidInternalStateError(
            f"Query plan has {len(head_nodes)} heads, expected exactly 1."
        )

    if head_node is None:
        head_node = plan[head_nodes[0]]

    # Special case handling for explain/show/set queries
    if isinstance(head_node, ExplainNode):
        return explain(
            plan, analyze=head_node.analyze, _format=head_node.format
        ), ResultType.TABULAR

    if isinstance(head_node, SetVariableNode):
        return head_node(None), ResultType.NON_TABULAR

    if isinstance(head_node, (ShowValueNode, ShowCreateNode)):
        return head_node(None, None), ResultType.TABULAR

    context = ExecutionContext(num_workers=4)

    def inner_execute(plan: PhysicalPlan) -> Generator:
        """Runs the plan in a hybrid execution model (serial + async)."""
        try:
            # Start processing scan nodes first
            pump_nodes = [
                (nid, node) for nid, node in plan.depth_first_search_flat() if node.is_scan
            ]
            for pump_nid, pump_instance in pump_nodes:
                for morsel in pump_instance(None, None):
                    if morsel is not None:
                        yield from process_node(plan, pump_nid, morsel, None, context)
                yield from process_node(plan, pump_nid, EOS, None, context)

            # Process async tasks until completion
            yield from context.process_async_tasks(plan)

        finally:
            context.shutdown()

    return inner_execute(plan), ResultType.TABULAR


def process_node(
    plan: PhysicalPlan,
    nid: str,
    morsel: pyarrow.Table,
    join_leg: Optional[str],
    context: ExecutionContext,
) -> Generator:
    """Processes a node in the execution plan."""
    node = plan[nid]

    if node.is_scan:
        # Process scan nodes normally
        for _, child, leg in plan.outgoing_edges(nid):
            results = process_node(plan, child, morsel, leg, context)
            yield from (result for result in results if result is not None)
    else:
        if node.is_stateless:
            # Submit stateless nodes for parallel processing
            context.submit_task(node, morsel, join_leg, nid)
            # Process any completed tasks
            yield from context.process_async_tasks(plan)
        else:
            # Process stateful nodes serially
            results = node(morsel, join_leg)
            if results is not None:
                for result in results:
                    if result is not None:
                        children = plan.outgoing_edges(nid)
                        if not children and result != EOS:
                            yield result
                        for _, child, leg in children:
                            yield from process_node(plan, child, result, leg, context)


def event_loop(plan: PhysicalPlan) -> Generator:
    """Event loop to process the plan and async tasks."""
    context = ExecutionContext(num_workers=4)
    try:
        print("loop")
        yield from process_node(plan, plan.get_entry_point(), None, None, context)
        yield from context.process_async_tasks(plan)
    finally:
        context.shutdown()



"""
This setup script builds and installs Opteryx by compiling both Cython extensions and a
Rust module.

It detects the operating system to set the correct compiler flags, organizes
include paths, processes dependencies, and reads in version and project metadata. Multiple
extension modules are defined for optimized data operations.
"""

import glob
import os
import platform
import sys
from sysconfig import get_config_var
from typing import Any
from typing import Dict

import numpy
from Cython.Build import cythonize
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup
from setuptools_rust import RustExtension

LIBRARY = "opteryx"


def is_mac():  # pragma: no cover
    return platform.system().lower() == "darwin"


def is_win():  # pragma: no cover
    return platform.system().lower() == "windows"


def is_linux():  # pragma: no cover
    return platform.system().lower() == "linux"


REQUESTED_COMMANDS = {arg.lower() for arg in sys.argv[1:] if arg and not arg.startswith("-")}
SHOULD_BUILD_EXTENSIONS = "clean" not in REQUESTED_COMMANDS

if not SHOULD_BUILD_EXTENSIONS:
    REQUESTED_DISPLAY = ", ".join(sorted(REQUESTED_COMMANDS)) or "<none>"
    print(
        f"\033[38;2;255;208;0mSkipping native extension build for command(s):\033[0m {REQUESTED_DISPLAY}"
    )

if SHOULD_BUILD_EXTENSIONS:
    CPP_COMPILE_FLAGS = ["-O3"]
    C_COMPILE_FLAGS = ["-O3"]
    if is_mac():
        CPP_COMPILE_FLAGS += ["-std=c++17"]
    elif is_win():
        CPP_COMPILE_FLAGS += ["/std:c++17"]
    else:
        CPP_COMPILE_FLAGS += ["-std=c++17", "-march=native", "-fvisibility=default"]
        C_COMPILE_FLAGS += ["-march=native", "-fvisibility=default"]

    COMMON_WARNING_SUPPRESSIONS = [
        "-Wno-unused-function",
        "-Wno-unreachable-code-fallthrough",
        "-Wno-sign-compare",
        "-Wno-integer-overflow",
        "-Wno-unused-command-line-argument",
    ]
    C_COMPILE_FLAGS.extend(COMMON_WARNING_SUPPRESSIONS)
    CPP_COMPILE_FLAGS.extend(COMMON_WARNING_SUPPRESSIONS)

    # Dynamically get the default include paths
    include_dirs = [numpy.get_include(), "src/cpp", "src/c"]

    # Get the C++ include directory
    includedir = get_config_var("INCLUDEDIR")
    if includedir:
        include_dirs.append(os.path.join(includedir, "c++", "v1"))

    # Get the Python include directory
    includepy = get_config_var("INCLUDEPY")
    if includepy:
        include_dirs.append(includepy)

    # Check if paths exist
    include_dirs = [p for p in include_dirs if os.path.exists(p)]

    print("\033[38;2;255;85;85mInclude paths:\033[0m", include_dirs)

    def rust_build(setup_kwargs: Dict[str, Any]) -> None:
        setup_kwargs.update(
            {
                "rust_extensions": [RustExtension("opteryx.compute", "Cargo.toml", debug=False)],
                "zip_safe": False,
            }
        )

    __author__ = "notset"
    __version__ = "notset"

    with open(f"{LIBRARY}/__version__.py", mode="r", encoding="UTF8") as v:
        vers = v.read()
    exec(vers)  # nosec

    RELEASE_CANDIDATE = "a" not in __version__ and "b" not in __version__
    COMPILER_DIRECTIVES = {"language_level": "3"}
    COMPILER_DIRECTIVES["profile"] = not RELEASE_CANDIDATE
    COMPILER_DIRECTIVES["linetrace"] = not RELEASE_CANDIDATE

    print(f"\033[38;2;255;85;85mBuilding Opteryx version:\033[0m {__version__}")
    print(f"\033[38;2;255;85;85mStatus:\033[0m (test)", "(rc)" if RELEASE_CANDIDATE else "")

    with open("README.md", mode="r", encoding="UTF8") as rm:
        long_description = rm.read()

    extensions = [
        Extension(
            name="opteryx.third_party.abseil.containers",
            sources=[
                "opteryx/third_party/abseil/containers.pyx",
                "third_party/abseil/absl/hash/internal/hash.cc",
                "third_party/abseil/absl/hash/internal/city.cc",
                "third_party/abseil/absl/container/internal/raw_hash_set.cc",
                "third_party/abseil/absl/hash/internal/low_level_hash.cc",
                "third_party/abseil/absl/base/internal/raw_logging.cc",
                "third_party/abseil/absl/base/internal/strerror.cc",
                "third_party/abseil/absl/base/internal/sysinfo.cc",
                "third_party/abseil/absl/base/internal/spinlock_wait.cc",
            ],
            include_dirs=include_dirs + ["third_party/abseil"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
            extra_link_args=["-Lthird_party/abseil"],
        ),
        Extension(
            name="opteryx.third_party.alantsd.base64",
            sources=["opteryx/third_party/alantsd/base64.pyx", "third_party/alantsd/base64.c"],
            include_dirs=include_dirs + ["third_party/alantsd"],
            extra_compile_args=C_COMPILE_FLAGS + ["-std=c99", "-DBASE64_IMPLEMENTATION"],
            extra_link_args=["-Lthird_party/alantsd"],
        ),
        Extension(
            name="opteryx.third_party.cyan4973.xxhash",
            sources=[
                "opteryx/third_party/cyan4973/xxhash.pyx",
                "third_party/cyan4973/xxhash.c",
            ],
            include_dirs=include_dirs + ["third_party/cyan4973"],
            extra_compile_args=C_COMPILE_FLAGS,
            extra_link_args=["-Lthird_party/cyan4973"],
        ),
        Extension(
            name="opteryx.third_party.fastfloat.fast_float",
            sources=[
                "opteryx/third_party/fastfloat/fast_float.pyx",
            ],
            include_dirs=include_dirs + ["third_party/fastfloat/fast_float"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
            extra_link_args=["-Lthird_party/fastfloat/fast_float"],
        ),
        Extension(
            name="opteryx.third_party.tktech.csimdjson",
            sources=[
                "opteryx/third_party/tktech/csimdjson.pyx",
                "third_party/tktech/simdjson/simdjson.cpp",
                "third_party/tktech/simdjson/util.cpp",
            ],
            include_dirs=include_dirs + ["third_party/tktech/simdjson"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.third_party.ulfjack.ryu",
            sources=[
                "opteryx/third_party/ulfjack/ryu.pyx",
                "third_party/ulfjack/ryu/d2fixed.c",
            ],
            include_dirs=include_dirs + ["third_party/ulfjack/ryu"],
            extra_compile_args=C_COMPILE_FLAGS,
            extra_link_args=["-Lthird_party/ulfjack/ryu"],
        ),
        Extension(
            name="opteryx.compiled.functions.strings",
            sources=["opteryx/compiled/functions/strings.pyx", "src/cpp/simd_search.cpp"],
            include_dirs=include_dirs,
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.functions.timestamp",
            sources=["opteryx/compiled/functions/timestamp.pyx"],
            extra_compile_args=C_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.functions.vectors",
            sources=["opteryx/compiled/functions/vectors.pyx"],
            include_dirs=include_dirs,
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.aggregations.count_distinct",
            sources=["opteryx/compiled/aggregations/count_distinct.pyx"],
            language="c++",
            include_dirs=include_dirs + ["third_party/abseil"],
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.hash_table",
            sources=["opteryx/compiled/structures/hash_table.pyx"],
            include_dirs=include_dirs,
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.bloom_filter",
            sources=["opteryx/compiled/structures/bloom_filter.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=C_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.buffers",
            sources=["opteryx/compiled/structures/buffers.pyx", "src/cpp/intbuffer.cpp"],
            include_dirs=include_dirs,
            language="c++",
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
            extra_compile_args=CPP_COMPILE_FLAGS + ["-Wall", "-shared"],
        ),
        Extension(
            name="opteryx.compiled.structures.lru_k",
            sources=["opteryx/compiled/structures/lru_k.pyx"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.memory_pool",
            sources=["opteryx/compiled/structures/memory_pool.pyx"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.relation_statistics",
            sources=["opteryx/compiled/structures/relation_statistics.pyx"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.node",
            sources=["opteryx/compiled/structures/node.pyx"],
            extra_compile_args=C_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.io.disk_reader",
            sources=["opteryx/compiled/io/disk_reader.pyx", "src/cpp/disk_io.cpp"],
            include_dirs=include_dirs + ["src/cpp"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.table_ops.distinct",
            sources=["opteryx/compiled/table_ops/distinct.pyx"],
            include_dirs=include_dirs + ["third_party/abseil"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.table_ops.hash_ops",
            sources=["opteryx/compiled/table_ops/hash_ops.pyx"],
            include_dirs=include_dirs + ["third_party/abseil"],
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.table_ops.null_avoidant_ops",
            sources=["opteryx/compiled/table_ops/null_avoidant_ops.pyx"],
            include_dirs=include_dirs,
            language="c++",
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.third_party.fuzzy",
            sources=["opteryx/third_party/fuzzy/soundex.pyx"],
            extra_compile_args=C_COMPILE_FLAGS,
        ),
        Extension(
            name="opteryx.compiled.structures.memory_view_stream",
            sources=["opteryx/compiled/structures/memory_view_stream.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=C_COMPILE_FLAGS,
        ),
    ]

    # Add SIMD support flags
    machine = platform.machine().lower()
    system = platform.system().lower()
    if machine.startswith("arm") and not machine.startswith("aarch64"):
        if system != "darwin":
            CPP_COMPILE_FLAGS.append("-mfpu=neon")
    elif "x86" in machine or "amd64" in machine:
        CPP_COMPILE_FLAGS.append("-mavx2")

    # Auto-generate list_ops.pyx to include all individual .pyx files in the folder
    # This ensures new files are automatically included when added
    list_ops_dir = "opteryx/compiled/list_ops"
    list_ops_file = os.path.join(list_ops_dir, "list_ops.pyx")

    # Find all .pyx files in the list_ops directory (excluding list_ops.pyx itself)
    pyx_files = sorted(
        [
            os.path.basename(f)
            for f in glob.glob(os.path.join(list_ops_dir, "*.pyx"))
            if os.path.basename(f) != "list_ops.pyx"
        ]
    )

    # Generate the list_ops.pyx file with include directives
    with open(list_ops_file, "w", encoding="UTF8") as f:
        f.write("""# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

\"\"\"
Auto-generated consolidated list_ops module.
This file is automatically generated by setup.py and includes all individual
list operation files in the list_ops directory.

DO NOT EDIT THIS FILE MANUALLY - it will be overwritten during build.
\"\"\"

""")

        # Add include directives for each .pyx file
        for pyx_file in pyx_files:
            f.write(f'include "{pyx_file}"\n')

    print(
        f"\033[38;2;189;147;249mAuto-generated list_ops.pyx with {len(pyx_files)} includes\033[0m"
    )

    # Auto-generate joins.pyx to include all individual .pyx files in the folder
    # This ensures new files are automatically included when added
    joins_dir = "opteryx/compiled/joins"
    joins_file = os.path.join(joins_dir, "joins.pyx")

    # Find all .pyx files in the joins directory (excluding joins.pyx itself)
    joins_pyx_files = sorted(
        [
            os.path.basename(f)
            for f in glob.glob(os.path.join(joins_dir, "*.pyx"))
            if os.path.basename(f) != "joins.pyx"
        ]
    )

    # Generate the joins.pyx file with include directives
    with open(joins_file, "w", encoding="UTF8") as f:
        f.write("""# cython: language_level=3
# cython: nonecheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: infer_types=True
# cython: wraparound=False
# cython: boundscheck=False

\"\"\"
Auto-generated consolidated joins module.
This file is automatically generated by setup.py and includes all individual
join files in the joins directory.

DO NOT EDIT THIS FILE MANUALLY - it will be overwritten during build.
\"\"\"

""")

        # Add include directives for each .pyx file
        for pyx_file in joins_pyx_files:
            f.write(f'include "{pyx_file}"\n')

    print(
        f"\033[38;2;189;147;249mAuto-generated joins.pyx with {len(joins_pyx_files)} includes\033[0m"
    )

    re2_sources = [
        "third_party/re2/re2/bitstate.cc",
        "third_party/re2/re2/compile.cc",
        "third_party/re2/re2/dfa.cc",
        "third_party/re2/re2/filtered_re2.cc",
        "third_party/re2/re2/mimics_pcre.cc",
        "third_party/re2/re2/nfa.cc",
        "third_party/re2/re2/onepass.cc",
        "third_party/re2/re2/parse.cc",
        "third_party/re2/re2/perl_groups.cc",
        "third_party/re2/re2/prefilter.cc",
        "third_party/re2/re2/prefilter_tree.cc",
        "third_party/re2/re2/prog.cc",
        "third_party/re2/re2/re2.cc",
        "third_party/re2/re2/regexp.cc",
        "third_party/re2/re2/set.cc",
        "third_party/re2/re2/simplify.cc",
        "third_party/re2/re2/stringpiece.cc",
        "third_party/re2/re2/tostring.cc",
        "third_party/re2/re2/unicode_casefold.cc",
        "third_party/re2/re2/unicode_groups.cc",
        "third_party/re2/util/rune.cc",
        "third_party/re2/util/strutil.cc",
    ]

    list_ops_link_args = []
    if not is_mac():
        list_ops_link_args.append("-lcrypto")
    if not is_win():
        list_ops_link_args.append("-pthread")

    extensions.append(
        Extension(
            name="opteryx.compiled.list_ops.function_definitions",
            sources=[list_ops_file, "src/cpp/simd_search.cpp"] + re2_sources,
            language="c++",
            include_dirs=include_dirs
            + [
                "third_party/abseil",
                "third_party/apache",
                "opteryx/third_party/apache",
                "third_party/re2",
            ],
            extra_compile_args=CPP_COMPILE_FLAGS,
            extra_link_args=list_ops_link_args,
        ),
    )

    extensions.append(
        Extension(
            name="opteryx.compiled.joins.join_definitions",
            sources=[joins_file, "src/cpp/join_kernels.cpp", "src/cpp/intbuffer.cpp"],
            language="c++",
            include_dirs=include_dirs + ["third_party/abseil", "third_party/fastfloat/fast_float"],
            extra_compile_args=CPP_COMPILE_FLAGS,
        ),
    )

    setup_config = {
        "name": LIBRARY,
        "version": __version__,
        "description": "Python SQL Query Engine",
        "long_description": long_description,
        "long_description_content_type": "text/markdown",
        "packages": find_packages(include=[LIBRARY, f"{LIBRARY}.*"]),
        "python_requires": ">=3.11",
        "url": "https://github.com/mabel-dev/opteryx/",
        "ext_modules": (
            cythonize(extensions, compiler_directives=COMPILER_DIRECTIVES)
            if SHOULD_BUILD_EXTENSIONS
            else []
        ),
        "entry_points": {
            "console_scripts": ["opteryx=opteryx.command:main"],
        },
        "package_data": {
            "": ["*.pyx", "*.pxd"],
        },
    }

    rust_build(setup_config)

    setup(**setup_config)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

from typing import Tuple

from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType


def get_nodes_of_type_from_logical_plan(plan: LogicalPlan, types: Tuple[LogicalPlanStepType]):
    matches = []
    for node in plan.nodes(True):
        if node[1].node_type in types:
            matches.append(node)
    return matches


class OptimizerContext:
    """Context object to carry state"""

    def __init__(self, tree: LogicalPlan):
        self.node_id = None
        self.parent_nid = None
        self.last_nid = None
        self.pre_optimized_tree = tree
        self.optimized_plan = LogicalPlan()

        self.seen_projections: int = 0
        self.seen_unions: int = 0
        self.seen_distincts: int = 0
        self.seen_projects_since_distinct: int = 0

        self.false_filters: list = []
        """We collect FILTER(FALSE) nodes for later rewriting"""

        self.collected_predicates: list = []
        """We collect predicates we should be able to push to reads and joins"""

        self.collected_identities: set = set()
        """We collect column identities so we can push column selection as close to the read as possible, including off to remote systems"""

        self.collected_distincts: list = []
        """We collect distincts to try to eliminate rows earlier"""

        self.collected_limits: list = []
        """We collect limits to to to eliminate rows earlier"""

        self.collected_joins = []
        """We collect joins to try to rewrite to inner (or filter joins)"""

        self.distincted_indentities: set = set()
        """The columns that implicitly exist in the plan because of a distinct"""

        self.bag = {}


class OptimizationStrategy:
    def __init__(self, statistics):
        self.statistics = statistics

    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        raise NotImplementedError(
            "Visit method must be implemented in OptimizationStrategy classes."
        )

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        raise NotImplementedError(
            "Complete method must be implemented in OptimizationStrategy classes."
        )

    def should_i_run(self, plan: LogicalPlan) -> bool:
        return True



# isort: skip

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Physical Execution Operators

This module contains all physical execution operators used by Opteryx's query engine.
Operators form a tree structure where each node processes data and passes results to
its parent node, implementing the iterator pattern for memory-efficient processing.

Operator Categories:

Data Sources:
- ReaderNode: Reads data from connectors (files, databases, etc.)
- AsyncReaderNode: Async version for improved I/O performance
- NullReaderNode: Returns empty table with correct schema (for contradictory predicates)
- FunctionDatasetNode: Generates data from function calls

Joins:
- InnerJoinNode: Standard inner joins
- OuterJoinNode: Left, right, and full outer joins
- CrossJoinNode: Cartesian product joins
- NestedLoopJoinNode: Nested loop algorithm for joins
- UnnestJoinNode: Specialized join for array unnesting

Filtering and Selection:
- FilterNode: Applies WHERE clause predicates
- FilterJoinNode: Optimized filter for join conditions
- ProjectionNode: Column selection and expression evaluation
- DistinctNode: Removes duplicate rows

Aggregation:
- AggregateNode: Standard aggregation (SUM, COUNT, etc.)
- AggregateAndGroupNode: GROUP BY with aggregation
- SimpleAggregateNode: Single-group aggregation optimization
- SimpleAggregateAndGroupNode: Optimized single-group with GROUP BY

Sorting and Limiting:
- SortNode: ORDER BY implementation using TimSort
- HeapSortNode: Heap-based sorting for TOP-N queries
- LimitNode: LIMIT/OFFSET clause implementation

Set Operations:
- UnionNode: UNION and UNION ALL operations

Control and Meta:
- ExitNode: Query execution termination
- ExplainNode: Query plan explanation
- ShowColumnsNode: SHOW COLUMNS implementation
- ShowCreateNode: SHOW CREATE VIEW implementation
- ShowValueNode: SHOW variable value display
- SetVariableNode: SET variable operations

Base Classes:
- BasePlanNode: Base class for all operators
- JoinNode: Base class for join operators

Operator Development:
1. Inherit from BasePlanNode or appropriate base class
2. Implement execute() method that processes data morsels
3. Handle memory management and resource cleanup
4. Support statistics collection for optimization
5. Add comprehensive tests

Example:
    class MyOperatorNode(BasePlanNode):
        def execute(self, morsel):
            # Process the data morsel
            processed_data = self.process_data(morsel)
            yield processed_data

Performance Considerations:
- Use vectorized operations with PyArrow when possible
- Implement memory pooling for large data processing
- Consider async operations for I/O bound operators
- Profile memory usage and optimize accordingly
"""

from .base_plan_node import BasePlanNode, JoinNode  # isort: skip

from .aggregate_and_group_node import AggregateAndGroupNode  # Group is always followed by aggregate
from .aggregate_node import AGGREGATORS
from .aggregate_node import AggregateNode  # aggregate data
from .async_read_node import AsyncReaderNode
from .null_reader_node import NullReaderNode  # empty table for contradictory predicates
from .simple_aggregate_node import SimpleAggregateNode  # aggregate data
from .simple_aggregate_and_group_node import SimpleAggregateAndGroupNode  # aggregate data

# from .build_statistics_node import BuildStatisticsNode  # Analyze Tables
from .cross_join_node import CrossJoinNode  # CROSS JOIN
from .unnest_join_node import UnnestJoinNode  # CROSS JOIN UNNEST
from .distinct_node import DistinctNode  # remove duplicate records
from .exit_node import ExitNode
from .explain_node import ExplainNode  # EXPLAIN queries
from .filter_join_node import FilterJoinNode  # filter unwanted rows
from .filter_node import FilterNode  # filter unwanted rows
from .function_dataset_node import FunctionDatasetNode  # Dataset Constructors
from .heap_sort_node import HeapSortNode  # Heap

# from .information_schema_node import InformationSchemaNode  # information_schema
from .inner_join_node import InnerJoinNode
from .nested_loop_join_node import NestedLoopJoinNode

from .limit_node import LimitNode  # select the first N records

from .outer_join_node import OuterJoinNode
from .projection_node import ProjectionNode  # remove unwanted columns including renames
from .read_node import ReaderNode
from .set_variable_node import SetVariableNode
from .show_columns_node import ShowColumnsNode  # column details
from .show_create_node import ShowCreateNode  # SHOW CREATE VIEW

# from .show_databases_node import ShowDatabasesNode  # SHOW DATABASES
# from .show_functions_node import ShowFunctionsNode  # supported functions
from .show_value_node import ShowValueNode  # display node for SHOW
from .sort_node import SortNode  # order by selected columns
from .union_node import UnionNode


__all__ = [
    "BasePlanNode",
    "JoinNode",
    "AggregateAndGroupNode",
    "AGGREGATORS",
    "AggregateNode",
    "AsyncReaderNode",
    "NullReaderNode",
    "SimpleAggregateNode",
    "SimpleAggregateAndGroupNode",
    "CrossJoinNode",
    "UnnestJoinNode",
    "DistinctNode",
    "ExitNode",
    "ExplainNode",
    "FilterJoinNode",
    "FilterNode",
    "FunctionDatasetNode",
    "HeapSortNode",
    "InnerJoinNode",
    "NestedLoopJoinNode",
    "LimitNode",
    "OuterJoinNode",
    "ProjectionNode",
    "ReaderNode",
    "SetVariableNode",
    "ShowColumnsNode",
    "ShowCreateNode",
    "ShowValueNode",
    "SortNode",
    "UnionNode",
    "is_aggregator",
    "aggregators",
]


def is_aggregator(name):
    return name in AGGREGATORS


def aggregators():
    return list(AGGREGATORS.keys())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Grouping Node

This is a SQL Query Execution Plan Node.

This is the grouping node, it is always followed by the aggregation node, but
the aggregation node doesn't need the grouping node.


"""

import time

import numpy
import pyarrow
from orso.types import OrsoTypes

from opteryx import EOS
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate_and_append
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import QueryProperties
from opteryx.operators.aggregate_node import build_aggregations
from opteryx.operators.aggregate_node import extract_evaluations
from opteryx.operators.aggregate_node import project

from . import BasePlanNode

CHUNK_SIZE = 65536


class AggregateAndGroupNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)
        self.groups = list(parameters["groups"])
        self.aggregates = list(parameters["aggregates"])
        projection = list(parameters["projection"])

        # we're going to preload some of the evaluation

        # Replace offset based GROUP BYs with their column
        self.groups = [
            (
                group
                if not (group.node_type == NodeType.LITERAL and group.type == OrsoTypes.INTEGER)
                else projection[group.value - 1]
            )
            for group in self.groups
        ]

        # get all the columns anywhere in the groups or aggregates
        all_identifiers = [
            node.schema_column.identity
            for node in get_all_nodes_of_type(
                self.groups + self.aggregates, select_nodes=(NodeType.IDENTIFIER,)
            )
        ]
        self.all_identifiers = list(dict.fromkeys(all_identifiers))

        # Get any functions we need to execute before aggregating
        self.evaluatable_nodes = extract_evaluations(self.aggregates)

        # get the aggregated groupings and functions
        self.group_by_columns = list({node.schema_column.identity for node in self.groups})
        self.column_map, self.aggregate_functions = build_aggregations(self.aggregates)

        self.buffer = []
        self.max_buffer_size = (
            250  # Buffer size before partial aggregation (kept for future parallelization)
        )
        self._partial_aggregated = False  # Track if we've done a partial aggregation
        self._disable_partial_agg = False  # Can disable if partial agg isn't helping

    @property
    def config(self):  # pragma: no cover
        from opteryx.managers.expression import format_expression

        return f"AGGREGATE ({', '.join(format_expression(col) for col in self.aggregates)}) GROUP BY ({', '.join(format_expression(col) for col in self.groups)})"

    @property
    def name(self):  # pragma: no cover
        return "Group By"

    def execute(self, morsel: pyarrow.Table, **kwargs):
        if morsel == EOS:
            if not self.buffer:
                yield EOS
                return

            # Do final aggregation if we have buffered data
            table = pyarrow.concat_tables(
                self.buffer,
                promote_options="permissive",
            )

            # If we've done partial aggregations, the aggregate functions need adjusting
            # because columns like "*" have been renamed to "*_count"
            if self._partial_aggregated:
                # Build new aggregate functions for re-aggregating partial results
                adjusted_aggs = []
                adjusted_column_map = {}

                for field_name, function, _count_options in self.aggregate_functions:
                    # For COUNT aggregates, the column is now named "*_count" and we need to SUM it
                    if function == "count":
                        renamed_field = f"{field_name}_count"
                        adjusted_aggs.append((renamed_field, "sum", None))
                        # The final column will be named "*_count_sum", need to track for renaming
                        for orig_name, mapped_name in self.column_map.items():
                            if mapped_name == f"{field_name}_count":
                                adjusted_column_map[orig_name] = f"{renamed_field}_sum"
                    # For other aggregates, we can re-aggregate with the same function
                    else:
                        renamed_field = f"{field_name}_{function}".replace("_hash_", "_")
                        # Some aggregates can be re-aggregated (sum, max, min)
                        if function in ("sum", "max", "min", "hash_one", "all", "any"):
                            adjusted_aggs.append((renamed_field, function, None))
                            # Track the mapping: original -> intermediate -> final
                            for orig_name, mapped_name in self.column_map.items():
                                if mapped_name == renamed_field:
                                    # sum->sum, max->max, etc. means same name
                                    adjusted_column_map[orig_name] = (
                                        f"{renamed_field}_{function}".replace("_hash_", "_")
                                    )
                        elif function == "mean":
                            # For mean, just take one of the existing values (not ideal)
                            adjusted_aggs.append((renamed_field, "hash_one", None))
                            for orig_name, mapped_name in self.column_map.items():
                                if mapped_name == renamed_field:
                                    adjusted_column_map[orig_name] = f"{renamed_field}_one"
                        elif function == "hash_list":
                            # For ARRAY_AGG, we need to flatten lists
                            adjusted_aggs.append((renamed_field, "hash_list", None))
                            for orig_name, mapped_name in self.column_map.items():
                                if mapped_name == renamed_field:
                                    adjusted_column_map[orig_name] = f"{renamed_field}_list"
                        else:
                            # For other aggregates, take one value
                            adjusted_aggs.append((renamed_field, "hash_one", None))
                            for orig_name, mapped_name in self.column_map.items():
                                if mapped_name == renamed_field:
                                    adjusted_column_map[orig_name] = f"{renamed_field}_one"

                groups = table.group_by(self.group_by_columns)
                groups = groups.aggregate(adjusted_aggs)

                # Use the adjusted column map for selecting/renaming
                groups = groups.select(list(adjusted_column_map.values()) + self.group_by_columns)
                groups = groups.rename_columns(
                    list(adjusted_column_map.keys()) + self.group_by_columns
                )
            else:
                groups = table.group_by(self.group_by_columns)
                groups = groups.aggregate(self.aggregate_functions)

                # project to the desired column names from the pyarrow names
                groups = groups.select(list(self.column_map.values()) + self.group_by_columns)
                groups = groups.rename_columns(list(self.column_map.keys()) + self.group_by_columns)

            # do the secondary activities for ARRAY_AGG (order and limit)
            array_agg_nodes = [
                node
                for node in get_all_nodes_of_type(
                    self.aggregates, select_nodes=(NodeType.AGGREGATOR,)
                )
                if node.value == "ARRAY_AGG" and (node.order or node.limit)
            ]

            if array_agg_nodes:
                # Process all ARRAY_AGG columns that need ordering/limiting
                arrays_to_update = {}
                field_defs = {}

                for node in array_agg_nodes:
                    column_name = node.schema_column.identity

                    # Store field definition before we drop the column
                    field_defs[column_name] = groups.field(column_name)

                    # Extract and process the data
                    column_data = groups.column(column_name).to_pylist()

                    # Apply ordering if specified
                    if node.order:
                        column_data = [
                            sorted(c, reverse=bool(node.order[0][1])) for c in column_data
                        ]

                    # Apply limit if specified
                    if node.limit:
                        column_data = [c[: node.limit] for c in column_data]

                    arrays_to_update[column_name] = column_data

                # Drop all columns we're updating
                columns_to_drop = list(arrays_to_update.keys())
                groups = groups.drop(columns_to_drop)

                # Append all updated columns back
                for column_name, column_data in arrays_to_update.items():
                    groups = groups.append_column(field_defs[column_name], [column_data])

            num_rows = groups.num_rows
            for start in range(0, num_rows, CHUNK_SIZE):
                yield groups.slice(start, min(CHUNK_SIZE, num_rows - start))

            yield EOS
            return

        morsel = project(morsel, self.all_identifiers)
        # Add a "*" column, this is an int because when a bool it miscounts
        # FIX: Use int8 as the comment states (bool can miscount)
        if "*" not in morsel.column_names:
            morsel = morsel.append_column(
                "*", [numpy.ones(shape=morsel.num_rows, dtype=numpy.int8)]
            )
        eval_start = time.monotonic_ns()
        if self.evaluatable_nodes:
            morsel = evaluate_and_append(self.evaluatable_nodes, morsel)
        morsel = evaluate_and_append(self.groups, morsel)
        self.statistics.time_group_by_evaluations += time.monotonic_ns() - eval_start

        self.buffer.append(morsel)

        # If buffer is full, do partial aggregation
        # BUT: Skip partial aggregation if it's not reducing data effectively
        if len(self.buffer) >= self.max_buffer_size and not self._disable_partial_agg:
            table = pyarrow.concat_tables(
                self.buffer,
                promote_options="permissive",
            )

            groups = table.group_by(self.group_by_columns)
            groups = groups.aggregate(self.aggregate_functions)

            # Check if partial aggregation is effective
            # If we're not reducing the row count significantly, stop doing partial aggs
            reduction_ratio = groups.num_rows / table.num_rows if table.num_rows > 0 else 1
            if reduction_ratio > 0.75:  # Kept more than 75% of rows - high cardinality!
                # Partial aggregation isn't helping, disable it and keep buffering
                self._disable_partial_agg = True
                # Don't replace buffer with partial result, keep accumulating
            else:
                # Good reduction, keep using partial aggregation
                self.buffer = [groups]  # Replace buffer with partial result
                self._partial_aggregated = True  # Mark that we've done a partial aggregation

        yield None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
This module provides the execution engine for processing physical plans in a serial manner.
"""

from typing import Any
from typing import Generator
from typing import Optional
from typing import Tuple

import pyarrow

from opteryx import EOS
from opteryx.constants import ResultType
from opteryx.exceptions import InvalidInternalStateError
from opteryx.models import PhysicalPlan
from opteryx.models import QueryStatistics


def execute(
    plan: PhysicalPlan, head_node: str = None, statistics: QueryStatistics = None
) -> Tuple[Generator[pyarrow.Table, Any, Any], ResultType]:
    from opteryx.operators import ExplainNode
    from opteryx.operators import SetVariableNode
    from opteryx.operators import ShowCreateNode
    from opteryx.operators import ShowValueNode

    # Retrieve the tail of the query plan, which should ideally be a single head node
    head_nodes = list(set(plan.get_exit_points()))

    if len(head_nodes) != 1:
        raise InvalidInternalStateError(
            f"Query plan has {len(head_nodes)} heads, expected exactly 1."
        )

    if head_node is None:
        head_node = plan[head_nodes[0]]

    # Special case handling for 'Explain' queries
    if isinstance(head_node, ExplainNode):
        return explain(
            plan, analyze=head_node.analyze, _format=head_node.format
        ), ResultType.TABULAR

    # Special case handling
    if isinstance(head_node, SetVariableNode):
        # Set the variables and return a non-tabular result
        return head_node(None), ResultType.NON_TABULAR
    if isinstance(head_node, (ShowValueNode, ShowCreateNode)):
        # There's no execution plan to execute, just return the result
        return head_node(None, None), ResultType.TABULAR

    def inner_execute(plan: PhysicalPlan) -> Generator:
        # Get the pump nodes from the plan and execute them in order
        pump_nodes = [(nid, node) for nid, node in plan.depth_first_search_flat() if node.is_scan]
        for pump_nid, pump_instance in pump_nodes:
            for morsel in pump_instance(None, None):
                if morsel is not None:
                    yield from process_node(plan, pump_nid, morsel, None)
            yield from process_node(plan, pump_nid, EOS, None)

    return inner_execute(plan), ResultType.TABULAR


def explain(
    plan: PhysicalPlan, analyze: bool, _format: str
) -> Generator[pyarrow.Table, None, None]:
    from opteryx import operators

    def _inner_explain(node, depth):
        incoming_operators = plan.ingoing_edges(node)
        for operator_name in incoming_operators:
            operator = plan[operator_name[0]]
            if isinstance(operator, (operators.ExitNode, operators.ExplainNode)):  # Skip ExitNode
                yield from _inner_explain(operator_name[0], depth)
                continue
            elif isinstance(operator, operators.BasePlanNode):
                record = {
                    "identity": operator.identity,
                    "tree": depth,
                    "operator": operator.name,
                    "config": operator.config,
                }
                if analyze:
                    record["time_ms"] = operator.execution_time / 1e6
                    record["records_in"] = operator.records_in
                    record["records_out"] = operator.records_out
                    record["bytes_in"] = operator.bytes_in
                    record["bytes_out"] = operator.bytes_out
                    record["calls"] = operator.calls
                yield record
                yield from _inner_explain(operator_name[0], depth + 1)

    head = list(dict.fromkeys(plan.get_exit_points()))
    if len(head) != 1:  # pragma: no cover
        raise InvalidInternalStateError(f"Problem with the plan - it has {len(head)} heads.")

    # for EXPLAIN ANALYZE, we execute the query and report statistics
    if analyze:
        # we don't want the results, just the details from the plan
        temp = None
        head_node = plan.get_exit_points()[0]
        query_head, _, _ = plan.ingoing_edges(head_node)[0]
        results, result_type = execute(plan, query_head)
        if results is not None:
            results_generator, _ = next(results, ([], None))
            for temp in results_generator:
                pass
        del temp

    explained = list(_inner_explain(head[0], 1))

    if _format == "TEXT":
        table = pyarrow.Table.from_pylist(explained).select(
            col for col in explained[0] if col not in ["identity", "bytes_in", "bytes_out"]
        )
    else:
        from opteryx.utils import mermaid

        mermaid_plan = mermaid.plan_to_mermaid(plan, explained)
        # DEBUG: print(mermaid_plan)
        table = pyarrow.Table.from_pylist([{"plan": mermaid_plan}])

    yield table


def process_node(
    plan: PhysicalPlan, nid: str, morsel: pyarrow.Table, join_leg: Optional[str]
) -> Generator:
    node = plan[nid]

    if node.is_scan:
        for _, child, leg in plan.outgoing_edges(nid):
            results = process_node(plan, child, morsel, leg)
            yield from (result for result in results if result is not None)
    else:
        results = node(morsel, join_leg)
        if results is None:
            yield None
            return
        for result in (result for result in results if result is not None):
            children = plan.outgoing_edges(nid)
            if len(children) == 0 and result != EOS:
                yield result
            for _, child, leg in children:
                yield from process_node(plan, child, result, leg)



# isort: skip_file
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Opteryx is a SQL query engine optimized for speed and efficiency.

To get started:
    import opteryx
    results = opteryx.query("SELECT * FROM my_table")

Opteryx handles parsing, planning, and execution of SQL queries with a focus on low-latency analytics over
local or remote data sources.

For more information check out https://opteryx.dev.
"""

import datetime
import os
import random
import time
import warnings
import platform
from pathlib import Path

from decimal import getcontext
from typing import Optional, Union, Dict, Any, List, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover - only for type checkers
    import pyarrow

# Set Decimal precision to 28 globally
getcontext().prec = 28

# end-of-stream marker
EOS: int = random.randint(-(2**63), 2**63 - 1)


def is_mac() -> bool:  # pragma: no cover
    """
    Check if the current platform is macOS.

    Returns:
        bool: True if the platform is macOS, False otherwise.
    """
    return platform.system().lower() == "darwin"


# we do a separate check for debug mode here so we don't load the config module just yet
OPTERYX_DEBUG = os.environ.get("OPTERYX_DEBUG") is not None


# python-dotenv allows us to create an environment file to store secrets.
# Only try to import dotenv if a .env file exists to avoid paying the
# import cost when no environment file is present.
_env_path = Path.cwd() / ".env"
if _env_path.exists():
    try:
        import dotenv  # type:ignore

        dotenv.load_dotenv(dotenv_path=_env_path)
        if OPTERYX_DEBUG:
            print(f"{datetime.datetime.now()} [LOADER] Loading `.env` file.")
    except ImportError:  # pragma: no cover
        # dotenv is optional; if it's not installed, just continue.
        pass


if OPTERYX_DEBUG:  # pragma: no cover
    from opteryx.debugging import OpteryxOrsoImportFinder

from opteryx import config
from opteryx.managers.cache.cache_manager import CacheManager  # isort:skip

_cache_manager = CacheManager(cache_backend=None)


def get_cache_manager() -> CacheManager:
    """Function to get the current cache manager."""
    return _cache_manager


# Lazy import registration functions - these are lightweight
from opteryx.connectors import register_arrow
from opteryx.connectors import register_df
from opteryx.connectors import register_store

from opteryx.__version__ import __author__
from opteryx.__version__ import __build__
from opteryx.__version__ import __version__


__all__ = [
    "apilevel",
    "connect",
    "Connection",
    "paramstyle",
    "query",
    "query_to_arrow",
    "register_arrow",
    "register_df",
    "register_store",
    "threadsafety",
    "__author__",
    "__build__",
    "__version__",
    "OPTERYX_DEBUG",
]

# PEP-249 specifies these attributes for a Python Database API 2.0 compliant interface
# For more details, see: https://www.python.org/dev/peps/pep-0249/
apilevel: str = "1.0"  # Compliance level with DB API 2.0
threadsafety: int = 0  # Thread safety level, 0 means not thread-safe
paramstyle: str = "named"  # Parameter placeholder style, named means :name for placeholders


def connect(*args, **kwargs) -> "Connection":
    """
    Establish a new database connection and return a Connection object.

    Note: This function is designed to comply with the 'connect' method
    described in PEP0249 for Python Database API Specification v2.0.
    """
    # Lazy import Connection
    from opteryx.connection import Connection

    # Check for deprecated 'cache' parameter
    if "cache" in kwargs:  # pragma: no cover
        # Import the warnings module here to minimize dependencies
        import warnings

        # Emit a deprecation warning
        warnings.warn(
            "'cache' is no longer set via a parameter on connect, use opteryx.cache_manager instead.",
            DeprecationWarning,
            stacklevel=2,
        )
    # Create and return a Connection object
    return Connection(*args, **kwargs)


def query(
    operation: str,
    params: Union[list, Dict, None] = None,
    visibility_filters: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    """
    Helper function to execute a query and return a cursor.

    This function is designed to be similar to the DuckDB function of the same name.
    It simplifies the process of executing queries by abstracting away the connection
    and cursor creation steps.

    Parameters:
        operation: SQL query string
        params: list of parameters to bind into the SQL query
        kwargs: additional arguments for creating the Connection

    Returns:
        Executed cursor
    """
    # Lazy import Connection
    from opteryx.connection import Connection

    # Create a new database connection
    conn = Connection(**kwargs)

    # Create a new cursor object using the connection
    curr = conn.cursor()

    # Execute the SQL query using the cursor
    curr.execute(operation=operation, params=params, visibility_filters=visibility_filters)

    # Return the executed cursor
    return curr


def query_to_arrow(
    operation: str,
    params: Union[List, Dict, None] = None,
    visibility_filters: Optional[Dict[str, Any]] = None,
    limit: int = None,
    **kwargs,
) -> "pyarrow.Table":
    """
    Helper function to execute a query and return a pyarrow Table.

    This is the fastest way to get a pyarrow table from Opteryx, it bypasses needing
    orso to create a Dataframe and converting from the Dataframe. This is fast, but
    not doing it is faster.

    Parameters:
        operation: SQL query string
        params: list of parameters to bind into the SQL query (optional)
        limit: stop after this many rows (optional)
        kwargs: additional arguments for creating the Connection

    Returns:
        pyarrow Table
    """
    # Lazy import Connection
    from opteryx.connection import Connection

    # Create a new database connection
    conn = Connection(**kwargs)

    # Create a new cursor object using the connection
    curr = conn.cursor()

    # Execute the SQL query using the cursor
    return curr.execute_to_arrow(
        operation=operation, params=params, visibility_filters=visibility_filters, limit=limit
    )


# Try to increase the priority of the application
if not config.DISABLE_HIGH_PRIORITY and hasattr(os, "nice"):  # pragma: no cover
    nice_value = os.nice(0)
    try:
        if not is_mac():
            os.nice(-20 + nice_value)
    except PermissionError:
        display_nice = str(nice_value)
        if nice_value == 0:
            display_nice = "0 (normal)"
        if OPTERYX_DEBUG:
            print(
                f"{datetime.datetime.now()} [LOADER] Cannot update process priority. Currently set to {display_nice}."
            )


def set_cache_manager(new_cache_manager):
    """Function to set a new cache manager and trigger custom functionality."""
    global _cache_manager
    _cache_manager = new_cache_manager

    # if we change the cache config, reset the BufferPool
    # Lazy import BufferPool
    from opteryx.shared import BufferPool

    BufferPool.reset()


cache_manager = get_cache_manager()


# if we're running in a notebook, register a magick
try:  # pragma: no cover
    from IPython import get_ipython
    from IPython.core.magic import Magics, magics_class, cell_magic

    @magics_class
    class OpteryxMagics(Magics):
        @cell_magic
        def opteryx(self, line, cell):
            self.shell.run_cell(
                'import opteryx\nopteryx.query("' + cell.replace("\n", " ") + '")',
                store_history=True,
            )

    ipython = get_ipython()
    if ipython:
        ipython.register_magics(OpteryxMagics)
except (ImportError, ValueError, TypeError) as err:  # pragma: no cover
    pass

# Enable all warnings, including DeprecationWarning
warnings.simplefilter("once", DeprecationWarning)

# Lazy initialization of system_statistics
_system_statistics = None


def _get_system_statistics():
    """
    Lazy getter for system statistics.

    System statistics are only created when first accessed, which avoids
    importing the QueryStatistics model (and its dependencies) during the
    initial import of the opteryx module.

    Returns:
        QueryStatistics: The global system statistics object
    """
    global _system_statistics
    if _system_statistics is None:
        from opteryx.models import QueryStatistics

        _system_statistics = QueryStatistics("system")
        _system_statistics.start_time = time.time_ns()
    return _system_statistics


# Provide access via module attribute
def __getattr__(name):
    """
    Lazy load module attributes to improve import performance.

    This function intercepts attribute access on the opteryx module to
    implement lazy loading of heavy components like Connection and
    system_statistics. This reduces initial import time from ~500ms to ~130ms.

    Supported lazy attributes:
    - Connection: The main connection class
    - system_statistics: Global query statistics object

    Args:
        name: The attribute name being accessed

    Returns:
        The requested attribute

    Raises:
        AttributeError: If the attribute doesn't exist
    """
    if name == "Connection":
        from opteryx.connection import Connection

        return Connection
    elif name == "system_statistics":
        return _get_system_statistics()
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Limit Node

This is a SQL Query Execution Plan Node.

This Node performs the LIMIT and the OFFSET steps
"""

import pyarrow

from opteryx import EOS
from opteryx.models import QueryProperties

from . import BasePlanNode


class LimitNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        BasePlanNode.__init__(self, properties=properties, **parameters)
        self.limit = parameters.get("limit", float("inf"))
        self.offset = parameters.get("offset", 0)

        self.remaining_rows = self.limit if self.limit is not None else float("inf")
        self.rows_left_to_skip = max(0, self.offset)

    @property
    def name(self):  # pragma: no cover
        return "LIMIT"

    @property
    def config(self):  # pragma: no cover
        return str(self.limit) + " OFFSET " + str(self.offset)

    def execute(self, morsel: pyarrow.Table, **kwargs) -> pyarrow.Table:
        if morsel == EOS:
            yield EOS
            return

        if self.rows_left_to_skip > 0:
            if self.rows_left_to_skip >= morsel.num_rows:
                self.rows_left_to_skip -= morsel.num_rows
                yield morsel.slice(offset=0, length=0)
                return
            else:
                morsel = morsel.slice(
                    offset=self.rows_left_to_skip, length=morsel.num_rows - self.rows_left_to_skip
                )
                self.rows_left_to_skip = 0

        if self.remaining_rows <= 0 or morsel.num_rows == 0:
            yield morsel.slice(offset=0, length=0)

        elif morsel.num_rows < self.remaining_rows:
            self.remaining_rows -= morsel.num_rows
            yield morsel
        else:
            rows_to_slice = self.remaining_rows
            self.remaining_rows = 0
            yield morsel.slice(offset=0, length=rows_to_slice)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Cross Join Node

This is a SQL Query Execution Plan Node.

This performs a CROSS JOIN - CROSS JOIN is not natively supported by PyArrow so this is written
here rather than calling the join() functions

Note: CROSS JOIN UNNEST is implemented by the UnnestJoinNode
"""

import numpy
import pyarrow

from opteryx import EOS
from opteryx.models import QueryProperties
from opteryx.third_party.abseil.containers import FlatHashSet

from . import JoinNode

INTERNAL_BATCH_SIZE: int = 10000  # config
MAX_JOIN_SIZE: int = 1_000_000  # config


def _cartesian_product(*arrays):
    """
    Cartesian product of arrays creates every combination of the elements in the arrays
    """
    array_count = len(arrays)
    arr = numpy.empty([len(array) for array in arrays] + [array_count], dtype=numpy.int64)
    for i, array in enumerate(numpy.ix_(*arrays)):
        arr[..., i] = array
    return numpy.hsplit(arr.reshape(-1, array_count), array_count)


def _cross_join(left_morsel, right):
    """
    A cross join is the cartesian product of two tables - this usually isn't very
    useful, but it does allow you to the theta joins (non-equi joins)
    """

    def _chunker(seq_1, seq_2, size: int = INTERNAL_BATCH_SIZE):
        for i in range(0, len(seq_1), size):
            yield memoryview(seq_1)[i : i + size], memoryview(seq_2)[i : i + size]

    from opteryx.utils.arrow import align_tables

    at_least_once = False
    left_schema = left_morsel.schema
    right_schema = right.schema

    # Iterate through left table in chunks of size INTERNAL_BATCH_SIZE
    for left_block in left_morsel.to_batches(max_chunksize=INTERNAL_BATCH_SIZE):
        # Convert the chunk to a table to retain column names
        left_block = pyarrow.Table.from_batches([left_block], schema=left_morsel.schema)

        # Create an array of row indices for each table
        left_array = numpy.arange(left_block.num_rows, dtype=numpy.int64)
        right_array = numpy.arange(right.num_rows, dtype=numpy.int64)

        # Calculate the cartesian product of the two arrays of row indices
        left_align, right_align = _cartesian_product(left_array, right_array)

        # Further break down the result into manageable chunks of size MAX_JOIN_SIZE
        for left_chunk, right_chunk in _chunker(
            left_align.flatten(), right_align.flatten(), MAX_JOIN_SIZE
        ):
            # Align the tables using the specified chunks of row indices
            table = align_tables(left_block, right, left_chunk, right_chunk)

            # Yield the resulting table to the caller
            yield table
            at_least_once = True

    if not at_least_once:
        fields = [pyarrow.field(name=f.name, type=f.type) for f in right_schema] + [
            pyarrow.field(name=f.name, type=f.type) for f in left_schema
        ]
        combined_schemas = pyarrow.schema(fields)
        yield pyarrow.Table.from_arrays(
            [pyarrow.array([]) for _ in combined_schemas], schema=combined_schemas
        )


class CrossJoinNode(JoinNode):
    """
    Implements a SQL CROSS JOIN
    """

    join_type = "cross"

    def __init__(self, properties: QueryProperties, **parameters):
        JoinNode.__init__(self, properties=properties, **parameters)

        self.source = parameters.get("column")

        self._left_relation = parameters.get("left_relation_names")
        self._right_relation = parameters.get("right_relation_names")

        self.left_buffer = []
        self.right_buffer = []
        self.left_relation = None
        self.right_relation = None
        self.hash_set = FlatHashSet()

        self.continue_executing = True

    @property
    def name(self):  # pragma: no cover
        return "Cross Join"

    @property
    def config(self):  # pragma: no cover
        return f"CROSS JOIN"

    def execute(self, morsel: pyarrow.Table, join_leg: str) -> pyarrow.Table:
        if not self.continue_executing:
            yield None
            return

        if join_leg == "left":
            if morsel == EOS:
                self.left_relation = pyarrow.concat_tables(self.left_buffer, promote_options="none")
                self.left_buffer.clear()
            else:
                self.left_buffer.append(morsel)
            yield None
            return

        if join_leg == "right":
            if morsel == EOS:
                right_table = pyarrow.concat_tables(self.right_buffer, promote_options="none")  # type:ignore
                self.right_buffer = None
                yield from _cross_join(self.left_relation, right_table)
                yield EOS
            else:
                self.right_buffer.append(morsel)
                yield None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Optimization Rule - Predicate rewriter

Type: Heuristic
Goal: Chose more efficient predicate evaluations

We rewrite conditions to a more optimal form based on two objectives:
1) the execution of the condition is faster
2) the condition is more likely to be able to be pushed to the storage layer (where its faster)

Rewrites Implemented:

x IN (single_value)                         → x = single_value
x NOT IN (single_value)                     → x != single_value
x LIKE 'pattern'                            → x = 'pattern' (when no wildcards)
x NOT LIKE 'pattern'                        → x != 'pattern' (when no wildcards)
x LIKE '%pattern%'                          → x INSTR 'pattern' (for contains without underscores)
x NOT LIKE '%pattern%'                      → x NOT INSTR 'pattern' (for contains without underscores)
x ILIKE '%pattern%'                         → x IINSTR 'pattern' (case-insensitive version)
x NOT ILIKE '%pattern%'                     → x NOT IINSTR 'pattern' (case-insensitive version)
x LIKE '%%%pattern%%'                       → x LIKE '%pattern%' (removing adjacent wildcards)
x ANY_OP = value                            → x IN (value) (when right side is a literal)
end - start > interval                      → start + interval < end (for date comparisons)
CASE WHEN x IS NULL THEN y ELSE x END       → IFNULL(x, y)
CASE WHEN x THEN y ELSE z END               → IIF(x, y, z)
COALESCE(x, y)                              → IFNULL(x, y) (when only two parameters)
SUBSTRING(x, 1, n)                          → LEFT(x, n) (when starting at position 1)
x LIKE 'pattern1%' OR x LIKE '%pattern2'    → x REGEX '^pattern1.*|.*pattern2$' (for ORed LIKE conditions)
CONCAT(x, y, z)                             → x || y || z (CONCAT to operators)
CONCAT_WS(x, y, z)                          → y || x || z (CONCAT_WS to operators)
x = 'a' OR x = 'b' OR x = 'c'               → x IN ('a', 'b', 'c') (for ORed Equals conditions)
a = ANY(z) OR b = ANY(z) OR c = ANY(z)      → (a, b, c) @> z
ENDS_WITH(x, pattern)                       → x LIKE '%pattern'
STARTS_WITH(x, pattern)                     → x LIKE 'pattern%'

#### IN THE PREDICATE ORDERING STRATEGY∂
a = ANY(z) AND b = ANY(z) AND c = ANY(z)    → z @>> (a, b, c)
"""

import re
from typing import Callable
from typing import Dict

from orso.schema import ConstantColumn
from orso.types import OrsoTypes

from opteryx.managers.expression import ExpressionColumn
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import format_expression
from opteryx.models import Node
from opteryx.models import QueryStatistics
from opteryx.planner.binder.operator_map import determine_type
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType
from opteryx.utils.sql import sql_like_to_regex

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext

# fmt: off
IN_REWRITES = {"InList": "Eq", "NotInList": "NotEq"}
LIKE_REWRITES = {"Like": "Eq", "NotLike": "NotEq"}
LITERALS_TO_THE_RIGHT = {"Plus": "Minus", "Minus": "Plus"}
INSTR_REWRITES = {"Like": "InStr", "NotLike": "NotInStr", "ILike": "IInStr", "NotILike": "NotIInStr"}
# fmt: on


def rewrite_in_to_eq(predicate):
    """
    Rewrite IN conditions with a single value to equality conditions.

    If the IN condition contains only one value, it is equivalent to an equality check.
    This optimization replaces the IN condition with a faster equality check.
    """
    predicate.value = IN_REWRITES[predicate.value]
    predicate.right.value = tuple(predicate.right.value)[0]
    predicate.right.type = predicate.right.element_type or OrsoTypes.VARCHAR
    predicate.right.element_type = None
    return predicate


def reorder_interval_calc(predicate):
    """
    rewrite:
        end - start > interval => start + interval > end

    This is because comparing a Date with a Date is faster than
    comparing in Interval with an Interval.
    """
    date_start = predicate.left.right
    date_end = predicate.left.left
    interval = predicate.right

    # Check if the operation is date - date
    if predicate.left.value == "Minus":
        # Create a new binary operator node for date + interval
        new_binary_op = Node(
            node_type=NodeType.BINARY_OPERATOR,
            value="Plus",
            left=date_start,
            right=interval,
        )
        binary_op_column_name = format_expression(new_binary_op, True)
        new_binary_op.schema_column = ExpressionColumn(
            name=binary_op_column_name, type=OrsoTypes.TIMESTAMP
        )

        # Create a new comparison operator node for date > date
        predicate.node_type = NodeType.COMPARISON_OPERATOR
        predicate.right = new_binary_op
        predicate.left = date_end

        predicate_column_name = format_expression(predicate, True)
        predicate.schema_column = ExpressionColumn(
            name=predicate_column_name, type=OrsoTypes.BOOLEAN
        )

        return predicate


def rewrite_ored_like_to_regex(predicate, statistics):
    """
    Rewrite multiple OR'ed LIKE conditions on the same column to a single regex pattern.

    Example:
    col LIKE 'pattern1%' OR col LIKE '%pattern2' OR col LIKE '%pattern3%'
    -->
    col REGEX '^pattern1.*|.*pattern2$|.*pattern3.*$'

    This optimization reduces multiple string pattern checks to a single regex evaluation.
    """
    # Collect LIKE conditions that can be combined
    like_conditions = {}

    def collect_likes(node, likes_dict):
        # Base case: LIKE/ILIKE condition
        if node.node_type == NodeType.COMPARISON_OPERATOR and node.value in {"Like", "ILike"}:
            # Only proceed if the right side is a literal
            if node.right.node_type == NodeType.LITERAL:
                # Get column identifier for grouping
                col_id = None
                if node.left.node_type == NodeType.IDENTIFIER:
                    col_id = node.left.schema_column.identity

                if col_id:
                    is_case_sensitive = node.value == "Like"
                    if col_id not in likes_dict:
                        likes_dict[col_id] = {
                            "patterns": [],
                            "nodes": [],
                        }

                    likes_dict[col_id]["patterns"].append(
                        sql_like_to_regex(
                            node.right.value, full_match=False, case_sensitive=is_case_sensitive
                        )
                    )
                    likes_dict[col_id]["nodes"].append(node)
            return

        # Recursive cases
        if node.node_type == NodeType.OR:
            collect_likes(node.left, likes_dict)
            collect_likes(node.right, likes_dict)

    collect_likes(predicate, like_conditions)

    for col_id, like_data in like_conditions.items():
        if len(like_data["patterns"]) > 1:
            statistics.optimization_predicate_rewriter_like_to_regex += 1
            # Create a new regex pattern
            regex_pattern = "|".join(pattern for pattern in like_data["patterns"])
            new_node = like_data["nodes"][0]
            new_node.value = "RLike"
            new_node.right.value = regex_pattern
            for node in like_data["nodes"][1:]:
                node.value = False
                node.node_type = NodeType.LITERAL
                node.type = OrsoTypes.BOOLEAN

    return predicate


def rewrite_ored_any_eq_to_contains(predicate, statistics):
    """
    Rewrite multiple OR'ed ANYOPEQ conditions on the same column to a single @> condition.

    Example:
    'a' = ANY(z) OR 'b' = ANY(z) OR 'c' = ANY(z)
    -->
    ('a', 'b', 'c') @> z

    This rewrite reduces many repeated ANY checks to a single containment operation.
    """
    anyeq_conditions = {}

    def collect_any_eq(node, grouped):
        if node.node_type == NodeType.COMPARISON_OPERATOR and node.value == "AnyOpEq":
            # Match only: literal = ANY(identifier)
            if (
                node.left.node_type == NodeType.LITERAL
                and node.right.node_type == NodeType.IDENTIFIER
            ):
                col_id = node.right.schema_column.identity
                if col_id not in grouped:
                    grouped[col_id] = {"values": [], "nodes": [], "column_node": node.right}
                grouped[col_id]["values"].append(node.left.value)
                grouped[col_id]["nodes"].append(node)
            return

        if node.node_type == NodeType.OR:
            collect_any_eq(node.left, grouped)
            collect_any_eq(node.right, grouped)

    collect_any_eq(predicate, anyeq_conditions)

    for data in anyeq_conditions.values():
        if len(data["values"]) > 1:
            statistics.optimization_predicate_rewriter_anyeq_to_contains += 1

            # Build new comparison node: ('a', 'b', 'c') @> z
            new_node = data["nodes"][0]

            new_node.left.value = set(data["values"])
            new_node.left.element_type = new_node.left.type
            new_node.left.type = OrsoTypes.ARRAY
            new_node.left.schema_column = ConstantColumn(
                name=new_node.left.name,
                type=OrsoTypes.ARRAY,
                element_type=new_node.left.element_type,
                value=new_node.left.value,
            )

            new_node.value = "AtArrow"
            new_node.node_type = NodeType.COMPARISON_OPERATOR
            new_node.right = data["column_node"]

            new_node.left, new_node.right = new_node.right, new_node.left  # Swap sides

            # Disable the remaining OR nodes
            for node in data["nodes"][1:]:
                node.node_type = NodeType.LITERAL
                node.type = OrsoTypes.BOOLEAN
                node.value = False

    return predicate


def rewrite_ored_eq_to_inlist(predicate, statistics):
    """
    Rewrite multiple OR'ed Equals conditions on the same column to a single regex pattern.

    Example:
    name = 'Earth' OR name = 'Mars' OR name = 'Venus'
    -->
    name IN ('Earth', 'Mars', 'Venus')
    """
    # Collect Equals conditions that can be combined
    eq_conditions = {}

    def collect_eqs(node, eqs_dict):
        # Base case: LIKE/ILIKE condition
        if node.node_type == NodeType.COMPARISON_OPERATOR and node.value in {"Eq"}:
            # Only proceed if the right side is a literal
            if node.right.node_type == NodeType.LITERAL:
                # Get column identifier for grouping
                col_id = None
                if node.left.node_type == NodeType.IDENTIFIER:
                    col_id = node.left.schema_column.identity

                if col_id:
                    if col_id not in eqs_dict:
                        eqs_dict[col_id] = {
                            "values": [],
                            "nodes": [],
                        }

                    eqs_dict[col_id]["values"].append(node.right.value)
                    eqs_dict[col_id]["nodes"].append(node)
            return

        # Recursive cases
        if node.node_type == NodeType.OR:
            collect_eqs(node.left, eqs_dict)
            collect_eqs(node.right, eqs_dict)

    collect_eqs(predicate, eq_conditions)

    for col_id, eq_data in eq_conditions.items():
        if len(eq_data["values"]) > 1:
            statistics.optimization_predicate_rewriter_eqs_to_list += 1
            # Create a new regex pattern
            new_node = eq_data["nodes"][0]
            new_node.value = "InList"
            new_node.right.value = set(eq_data["values"])
            new_node.right.element_type = new_node.right.type
            new_node.right.type = OrsoTypes.ARRAY
            new_node.right.schema_column = ConstantColumn(
                name=new_node.right.name,
                type=OrsoTypes.ARRAY,
                element_type=new_node.right.element_type,
                value=new_node.right.value,
            )
            for node in eq_data["nodes"][1:]:
                node.value = False
                node.node_type = NodeType.LITERAL
                node.type = OrsoTypes.BOOLEAN

    return predicate


# Define dispatcher conditions and actions
dispatcher: Dict[str, Callable] = {
    "rewrite_in_to_eq": rewrite_in_to_eq,
    "reorder_interval_calc": reorder_interval_calc,
}


# Dispatcher conditions
def _rewrite_predicate(predicate, statistics: QueryStatistics):
    if predicate.node_type == NodeType.FUNCTION:
        return _rewrite_function(predicate, statistics)

    # Add our new rewrite for ORed LIKE conditions
    if predicate.node_type == NodeType.OR:
        rewritten = rewrite_ored_like_to_regex(predicate, statistics)
        rewritten = rewrite_ored_eq_to_inlist(rewritten, statistics)
        rewritten = rewrite_ored_any_eq_to_contains(rewritten, statistics)
        if rewritten != predicate:
            return rewritten

    # if predicate.node_type in {NodeType.AND, NodeType.OR, NodeType.XOR}:
    if predicate.left:
        predicate.left = _rewrite_predicate(predicate.left, statistics)
    if predicate.right:
        predicate.right = _rewrite_predicate(predicate.right, statistics)
    if predicate.centre:
        predicate.centre = _rewrite_predicate(predicate.centre, statistics)

    if predicate.node_type not in {NodeType.BINARY_OPERATOR, NodeType.COMPARISON_OPERATOR}:
        # after rewrites, some filters aren't actually predicates
        return predicate

    if predicate.right.type == OrsoTypes.VARCHAR:
        if predicate.value in {"Like", "ILike", "NotLike", "NotILike"}:
            if "%%" in predicate.right.value:
                statistics.optimization_predicate_rewriter_remove_adjacent_wildcards += 1
                predicate.right.value = re.sub(r"%+", "%", predicate.right.value)

        if predicate.value in LIKE_REWRITES:
            if "%" not in predicate.right.value and "_" not in predicate.right.value:
                statistics.optimization_predicate_rewriter_remove_redundant_like += 1
                predicate.value = LIKE_REWRITES[predicate.value]

        if predicate.value in INSTR_REWRITES:
            if (
                "_" not in predicate.right.value
                and predicate.right.value.endswith("%")
                and predicate.right.value.startswith("%")
                and "%" not in predicate.right.value[1:-1]
            ):
                statistics.optimization_predicate_rewriter_replace_like_with_in_string += 1
                predicate.right.value = predicate.right.value[1:-1]
                predicate.value = INSTR_REWRITES[predicate.value]

    if predicate.right.type == OrsoTypes.BLOB:
        if predicate.value in {"Like", "ILike", "NotLike", "NotILike"}:
            if b"%%" in predicate.right.value:
                statistics.optimization_predicate_rewriter_remove_adjacent_wildcards += 1
                predicate.right.value = re.sub(b"%+", b"%", predicate.right.value)

        if predicate.value in LIKE_REWRITES:
            if b"%" not in predicate.right.value and b"_" not in predicate.right.value:
                statistics.optimization_predicate_rewriter_remove_redundant_like += 1
                predicate.value = LIKE_REWRITES[predicate.value]

        if predicate.value in INSTR_REWRITES:
            if (
                b"_" not in predicate.right.value
                and predicate.right.value.endswith(b"%")
                and predicate.right.value.startswith(b"%")
            ):
                statistics.optimization_predicate_rewriter_replace_like_with_in_string += 1
                predicate.right.value = predicate.right.value[1:-1]
                predicate.value = INSTR_REWRITES[predicate.value]

    if predicate.value == "AnyOpEq":
        if predicate.right.node_type == NodeType.LITERAL:
            statistics.optimization_predicate_rewriter_any_to_inlist += 1
            predicate.value = "InList"

    if predicate.value == "AnyOpNotEq":
        if predicate.right.node_type == NodeType.LITERAL:
            statistics.optimization_predicate_rewriter_any_to_inlist += 1
            predicate.value = "NotInList"

    if predicate.value in IN_REWRITES:
        if predicate.right.node_type == NodeType.LITERAL and len(predicate.right.value) == 1:
            statistics.optimization_predicate_rewriter_in_to_equals += 1
            return dispatcher["rewrite_in_to_eq"](predicate)

    if (
        predicate.node_type == NodeType.COMPARISON_OPERATOR
        and predicate.left.node_type == NodeType.BINARY_OPERATOR
    ):
        if (
            determine_type(predicate.left) == OrsoTypes.INTERVAL
            and determine_type(predicate.right) == OrsoTypes.INTERVAL
        ):
            statistics.optimization_predicate_rewriter_date_ += 1
            predicate = dispatcher["reorder_interval_calc"](predicate)

    return predicate


def _rewrite_function(function, statistics: QueryStatistics):
    if function.value == "CASE":
        # CASE WHEN x IS NULL THEN y ELSE x END → IFNULL(x, y)
        if len(function.parameters) == 2 and function.parameters[0].parameters[0].value == "IsNull":
            compare_column = function.parameters[0].parameters[0].centre
            target_column = function.parameters[1].parameters[1]
            value_if_null = function.parameters[1].parameters[0]

            if compare_column.schema_column.identity == target_column.schema_column.identity:
                statistics.optimization_predicate_rewriter_case_to_ifnull += 1
                function.value = "IFNULL"
                function.parameters = [compare_column, value_if_null]
                return function
        # CASE WHEN x THEN y ELSE z END → IIF(x, y, z)
        if (
            len(function.parameters) == 2
            and len(function.parameters[0].parameters) == 2
            and function.parameters[0].parameters[1].value is True
            and len(function.parameters[1].parameters) == 2
        ):
            statistics.optimization_predicate_rewriter_case_to_iif += 1

            compare_column = function.parameters[0].parameters[0]
            value_if_true = function.parameters[1].parameters[0]
            value_if_false = function.parameters[1].parameters[1]

            function.value = "IIF"
            function.parameters = [compare_column, value_if_true, value_if_false]
            return function
    # COALESCE(x, y) → IFNULL(x, y)
    if function.value == "COALESCE":
        if len(function.parameters) == 2:
            statistics.optimization_predicate_rewriter_coalesce_to_ifnull += 1
            function.value = "IFNULL"
            return function
    # SUBSTRING(x, 1, n) → LEFT(x, n)
    if function.value == "SUBSTRING" and function.parameters[1].value == 1:
        statistics.optimization_predicate_rewriter_substring_to_left += 1
        function.value = "LEFT"
        function.parameters = [function.parameters[0], function.parameters[2]]
        return function
    # CONCAT(x, y, z) → x || y || z
    if function.value == "CONCAT" and len(function.parameters) > 1:
        statistics.optimization_predicate_rewriter_concat_to_double_pipe += 1
        left_node = function.parameters[0]
        for param in function.parameters[1:]:
            this_node = Node(
                node_type=NodeType.BINARY_OPERATOR,
                value="StringConcat",
                left=left_node,
                right=param,
                schema_column=ExpressionColumn(name="", type=OrsoTypes.VARCHAR),
            )
            left_node = this_node
        this_node.alias = function.alias
        this_node.schema_column = function.schema_column
        function = this_node
    # CONCAT_WS(x, y, z) → y || x || z
    if function.value == "CONCAT_WS" and len(function.parameters) > 2:
        statistics.optimization_predicate_rewriter_concatws_to_double_pipe += 1
        separator = function.parameters[0]
        left_node = function.parameters[1]
        for param in function.parameters[2:]:
            separator_node = Node(
                node_type=NodeType.BINARY_OPERATOR,
                value="StringConcat",
                left=left_node,
                right=separator,
                schema_column=ExpressionColumn(name="", type=OrsoTypes.VARCHAR),
            )
            this_node = Node(
                node_type=NodeType.BINARY_OPERATOR,
                value="StringConcat",
                left=separator_node,
                right=param,
                schema_column=ExpressionColumn(name="", type=OrsoTypes.VARCHAR),
            )
            left_node = this_node
        this_node.alias = function.alias
        this_node.schema_column = function.schema_column
        function = this_node

    # STARTS_WITH → x LIKE 'pattern%'
    if function.value == "STARTS_WITH":
        statistics.optimization_predicate_rewriter_starts_with_to_like += 1
        left_node = function.parameters[0]
        function.parameters[1].value = function.parameters[1].value + "%"
        this_node = Node(
            node_type=NodeType.COMPARISON_OPERATOR,
            value="Like",
            left=left_node,
            right=function.parameters[1],
            schema_column=ExpressionColumn(name="", type=OrsoTypes.BOOLEAN),
        )
        this_node.alias = function.alias
        this_node.schema_column = function.schema_column
        function = this_node

    # ENDS_WITH → x LIKE '%pattern'
    if function.value == "ENDS_WITH":
        statistics.optimization_predicate_rewriter_ends_with_to_like += 1
        left_node = function.parameters[0]
        function.parameters[1].value = "%" + function.parameters[1].value
        this_node = Node(
            node_type=NodeType.COMPARISON_OPERATOR,
            value="Like",
            left=left_node,
            right=function.parameters[1],
            schema_column=ExpressionColumn(name="", type=OrsoTypes.BOOLEAN),
        )
        this_node.alias = function.alias
        this_node.schema_column = function.schema_column
        function = this_node

    return function


class PredicateRewriteStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Filter:
            node.condition = _rewrite_predicate(node.condition, self.statistics)
            context.optimized_plan[context.node_id] = node

        if node.node_type == LogicalPlanStepType.Project:
            new_columns = []
            for column in node.columns:
                new_column = _rewrite_predicate(column, self.statistics)
                new_columns.append(new_column)
            node.columns = new_columns
            context.optimized_plan[context.node_id] = node

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        return plan



"""
Code was originally taken from
https://github.com/TomScheffers/pyarrow_ops/blob/main/pyarrow_ops/ops.py
and has been extensively modified for Opteryx.
"""

import re
from contextlib import suppress

import numpy
import pyarrow
from orso.types import OrsoTypes
from pyarrow import compute

from opteryx.compiled import list_ops

# Operators where null compression isn't safe
skip_compression_ops = {
    "InList",
    "NotInList",
    "AnyOpEq",
    "AnyOpNotEq",
    "AnyOpGt",
    "AnyOpGtEq",
    "AnyOpLt",
    "AnyOpLtEq",
    "AnyOpLike",
    "AnyOpNotLike",
    "AnyOpILike",
    "AnyOpNotILike",
    "AnyOpRLike",
    "AnyOpNotRLike",
    "AllOpEq",
    "AllOpNotEq",
    "AtArrow",
    "ArrayContainsAll",
}


def filter_operations(left_arr, left_type, operator, right_arr, right_type):
    """
    Wrapped for Opteryx added to correctly handle null semantics.

    This returns an array with tri-state boolean (tue/false/none);
    if being used for display use as is, if being used for filtering, none is false.
    """
    if len(left_arr) == 0 or len(right_arr) == 0:
        return numpy.empty(0, dtype=bool)

    # INTEGERS and DECIMALS don't play nicely so we cast the INTS to DOUBLES
    if left_type == OrsoTypes.DECIMAL and right_type == OrsoTypes.INTEGER:
        right_arr = compute.cast(right_arr, pyarrow.float64())
        right_type = OrsoTypes.DOUBLE
    elif right_type == OrsoTypes.DECIMAL and left_type == OrsoTypes.INTEGER:
        left_arr = compute.cast(left_arr, pyarrow.float64())
        left_type = OrsoTypes.DOUBLE

    morsel_size = len(left_arr)
    compressed = False

    if operator not in skip_compression_ops:
        # compressing ARRAY columns is VERY SLOW

        # compute null positions
        left_null_positions = compute.is_null(left_arr, nan_is_null=True)

        # if the right side is an array, combine the null positions
        if len(right_arr) > 1:
            right_null_positions = compute.is_null(right_arr, nan_is_null=True)
            null_positions = numpy.logical_or(left_null_positions, right_null_positions)
        # if the right side is a scalar and is null, we can just return all nulls
        elif len(right_arr) == 1 and right_arr[0] is None:
            return pyarrow.nulls(morsel_size, type=pyarrow.bool_())
        # if the right side is a scalar and is not null, we can just use the left nulls
        else:
            null_positions = left_null_positions.to_numpy(False)

        # Early exit if all values are null
        if null_positions.all():
            return pyarrow.nulls(morsel_size, type=pyarrow.bool_())

        if null_positions.any():
            # if we have nulls and both columns are numpy arrays, we can speed things
            # up by removing the nulls from the calculations, we add the rows back in
            # later
            valid_positions = ~null_positions
            compressed = True

            if isinstance(left_arr, numpy.ndarray):
                left_arr = left_arr.compress(valid_positions)
            else:
                left_arr = compute.filter(left_arr, valid_positions)

            if len(right_arr) > 1:
                if isinstance(right_arr, numpy.ndarray):
                    right_arr = right_arr.compress(valid_positions)
                else:
                    right_arr = compute.filter(right_arr, valid_positions)

    if (
        OrsoTypes.TIMESTAMP in (left_type, right_type) or OrsoTypes.DATE in (left_type, right_type)
    ) and OrsoTypes.INTEGER in (left_type, right_type):
        from opteryx.functions.date_functions import convert_int64_array_to_pyarrow_datetime

        if left_type == OrsoTypes.INTEGER:
            left_arr = convert_int64_array_to_pyarrow_datetime(left_arr)
            left_type = OrsoTypes.TIMESTAMP
        if right_type == OrsoTypes.INTEGER:
            right_arr = convert_int64_array_to_pyarrow_datetime(right_arr)
            right_type = OrsoTypes.TIMESTAMP

    if OrsoTypes.INTERVAL in (left_type, right_type):
        from opteryx.datatypes.intervals import INTERVAL_KERNELS

        function = INTERVAL_KERNELS.get((left_type, right_type, operator))
        if function is None:
            from opteryx.exceptions import UnsupportedTypeError

            raise UnsupportedTypeError(
                f"Cannot perform {operator.upper()} on {left_type} and {right_type}."
            )

        results_mask = function(left_arr, left_type, right_arr, right_type, operator)
    else:
        # do the evaluation
        results_mask = _inner_filter_operations(left_arr, operator, right_arr)

    if compressed:
        # build tri-state response, PyArrow supports tristate, numpy does not
        full_result = numpy.full(morsel_size, None, dtype=object)
        numpy.place(full_result, valid_positions, results_mask)
        return pyarrow.array(full_result, type=pyarrow.bool_())

    return results_mask


# Filter functionality
def _inner_filter_operations(arr, operator, value):
    """
    Execute filter operations, this returns an array of the indexes of the rows that
    match the filter
    """
    if not operator.startswith(("AnyOp", "AllOp")) and len(value) == 1:
        value = value[0]
        with suppress(AttributeError):
            value = value.item()
        if isinstance(value, (tuple, list)):
            value = pyarrow.array(value)

    if operator == "Eq":
        return compute.equal(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "NotEq":
        return compute.not_equal(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "Lt":
        return compute.less(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "Gt":
        return compute.greater(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "LtEq":
        return compute.less_equal(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "GtEq":
        return compute.greater_equal(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "InList":
        to_pylist = getattr(value, "to_pylist", None)
        if to_pylist is not None:
            value = to_pylist()

        to_numpy = getattr(value, "to_numpy", None)
        if to_numpy is not None:
            value = to_numpy(zero_copy_only=False)

        values = set(value)
        to_numpy = getattr(arr, "to_numpy", None)
        if to_numpy is not None:
            arr = to_numpy(zero_copy_only=False)
        if arr.dtype == numpy.int64:
            return list_ops.list_in_list_int64(memoryview(arr), values, len(arr))
        else:
            return list_ops.list_in_list(arr.astype(object), values)
    if operator == "NotInList":
        to_pylist = getattr(value, "to_pylist", None)
        if to_pylist is not None:
            value = to_pylist()

        to_numpy = getattr(value, "to_numpy", None)
        if to_numpy is not None:
            value = to_numpy(zero_copy_only=False)

        values = set(value)

        to_numpy = getattr(arr, "to_numpy", None)
        if to_numpy is not None:
            arr = to_numpy(zero_copy_only=False)

        if arr.dtype == numpy.int64:
            matches = list_ops.list_in_list_int64(memoryview(arr), values, len(arr))
        else:
            matches = list_ops.list_in_list(arr.astype(object), values)
        return numpy.invert(matches.astype(dtype=numpy.bool_))
    if operator == "InStr":
        needle = str(value)
        if not isinstance(arr, (pyarrow.Array, pyarrow.ChunkedArray)):
            arr = pyarrow.array(arr, type=pyarrow.binary())
        matches = list_ops.list_in_string(arr, needle)
        return numpy.frombuffer(matches, dtype=numpy.bool_)
    if operator == "NotInStr":
        needle = str(value)
        if not isinstance(arr, (pyarrow.Array, pyarrow.ChunkedArray)):
            arr = pyarrow.array(arr, type=pyarrow.binary())
        matches = list_ops.list_in_string(arr, needle)
        matches = numpy.frombuffer(matches, dtype=numpy.bool_)
        return numpy.invert(matches)
    if operator == "IInStr":
        needle = str(value)
        if not isinstance(arr, (pyarrow.Array, pyarrow.ChunkedArray)):
            arr = pyarrow.array(arr, type=pyarrow.binary())
        matches = list_ops.list_in_string_case_insensitive(arr, needle)
        return numpy.frombuffer(matches, dtype=numpy.bool_)
    if operator == "NotIInStr":
        needle = str(value)
        if not isinstance(arr, (pyarrow.Array, pyarrow.ChunkedArray)):
            arr = pyarrow.array(arr, type=pyarrow.binary())
        matches = list_ops.list_in_string_case_insensitive(arr, needle)
        matches = numpy.frombuffer(matches, dtype=numpy.bool_)
        return numpy.invert(matches)
    if operator == "Like":
        return compute.match_like(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "NotLike":
        matches = compute.match_like(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
        return numpy.invert(matches)
    if operator == "ILike":
        return (
            compute.match_like(arr, value, ignore_case=True)
            .to_numpy(False)
            .astype(dtype=numpy.bool_)
        )
    if operator == "NotILike":
        matches = compute.match_like(arr, value, ignore_case=True)
        return numpy.invert(matches)
    if operator == "RLike":
        return compute.match_substring_regex(arr, value).to_numpy(False).astype(dtype=numpy.bool_)
    if operator == "NotRLike":
        matches = compute.match_substring_regex(arr, value)  # [#325]
        return numpy.invert(matches)
    if operator == "AnyOpEq":
        return list_ops.list_anyop_eq(literal=arr[0], column=value)
    if operator == "AnyOpNotEq":
        return list_ops.list_anyop_neq(literal=arr[0], column=value)
    if operator == "AnyOpGt":
        return list_ops.list_anyop_gt(arr[0], value)
    if operator == "AnyOpLt":
        return list_ops.list_anyop_lt(arr[0], value)
    if operator == "AnyOpGtEq":
        return list_ops.list_anyop_gte(arr[0], value)
    if operator == "AnyOpLtEq":
        return list_ops.list_anyop_lte(arr[0], value)
    if operator == "AllOpEq":
        return list_ops.list_allop_eq(arr[0], value)
    if operator == "AllOpNotEq":
        return list_ops.list_allop_neq(arr[0], value)

    if operator == "AnyOpILike":
        from opteryx.utils.sql import regex_match_any

        return regex_match_any(arr, value, flags=re.IGNORECASE)

    if operator == "AnyOpLike":
        from opteryx.utils.sql import regex_match_any

        return regex_match_any(arr, value)

    if operator == "AnyOpNotLike":
        from opteryx.utils.sql import regex_match_any

        return regex_match_any(arr, value, invert=True)

    if operator == "AnyOpNotILike":
        from opteryx.utils.sql import regex_match_any

        return regex_match_any(arr, value, flags=re.IGNORECASE, invert=True)

    if operator == "AtQuestion":
        from opteryx.third_party.tktech import csimdjson as simdjson

        to_numpy = getattr(arr, "to_numpy", None)
        if to_numpy is not None:
            arr = to_numpy(zero_copy_only=False)

        parser = simdjson.Parser()

        if not value.startswith("$."):
            # Not a JSONPath, treat as a simple key existence check
            return pyarrow.array(
                [value in parser.parse(doc).keys() for doc in arr],
                type=pyarrow.bool_(),  # type: ignore
            )

        # Convert "$.key1.list[0]" to JSON Pointer "/key1/list/0"
        def jsonpath_to_pointer(jsonpath: str) -> str:
            # Remove "$." prefix
            json_pointer = jsonpath[1:]
            # Replace "." with "/" for dict navigation
            json_pointer = json_pointer.replace(".", "/")
            # Replace "[index]" with "/index" for list access
            json_pointer = json_pointer.replace("[", "/").replace("]", "")
            return json_pointer

        # Convert "$.key1.key2" to JSON Pointer "/key1/key2"
        json_pointer = jsonpath_to_pointer(value)

        def check_json_pointer(doc, pointer):
            try:
                # Try accessing the path via JSON Pointer
                parser.parse(doc).at_pointer(pointer)
                return True  # If successful, the path exists
            except Exception:
                return False  # If an error occurs, the path does not exist

        # Apply the JSON Pointer check
        return pyarrow.array(
            [check_json_pointer(doc, json_pointer) for doc in arr],
            type=pyarrow.bool_(),
        )

    if operator == "AtArrow":
        from opteryx.compiled.list_ops import list_contains_any

        to_pylist = getattr(value, "to_pylist", None)
        if to_pylist is not None:
            value = to_pylist()

        to_numpy = getattr(arr, "to_numpy", None)
        if to_numpy is not None:
            arr = to_numpy(zero_copy_only=False)

        if len(arr) == 0:
            return numpy.array([], dtype=numpy.bool_)
        if len(arr) == 1:
            return numpy.array([set(arr[0]).intersection(value)], dtype=numpy.bool_)

        return list_contains_any(arr, set(value))

    if operator == "ArrayContainsAll":
        from opteryx.compiled.list_ops import list_contains_all

        to_pylist = getattr(value, "to_pylist", None)
        if to_pylist is not None:
            value = to_pylist()

        to_numpy = getattr(arr, "to_numpy", None)
        if to_numpy is not None:
            arr = to_numpy(zero_copy_only=False)

        if len(arr) == 1 and len(value) != 0:
            raise ValueError("Unable to execute @>>, check form matches `column @>> (values)`.")

        return list_contains_all(arr, set(value))

    raise NotImplementedError(f"Operator {operator} is not implemented!")  # pragma: no cover



from .cache_manager import CacheManager


# Only import the cache manager being used
def get_cache_manager(cache_type: str):
    """Return the cache manager class for the requested cache_type.

    This performs the import lazily so importing this package is cheaper when
    a backend is not needed.
    """
    if cache_type == "memcached":
        from .memcached import MemcachedCache

        return MemcachedCache
    elif cache_type == "redis":
        from .redis import RedisCache

        return RedisCache
    elif cache_type == "valkey":
        from .valkey import ValkeyCache

        return ValkeyCache
    else:
        from .null_cache import NullCache

        return NullCache


def __getattr__(name: str):
    """Lazily expose concrete cache manager classes as module attributes.

    This keeps `from opteryx.managers.cache import MemcachedCache` working
    while still avoiding eager imports of all backends.
    """
    if name == "MemcachedCache":
        from .memcached import MemcachedCache

        return MemcachedCache
    if name == "RedisCache":
        from .redis import RedisCache

        return RedisCache
    if name == "ValkeyCache":
        from .valkey import ValkeyCache

        return ValkeyCache
    if name == "NullCache":
        from .null_cache import NullCache

        return NullCache
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = (
    "CacheManager",
    "get_cache_manager",
    "MemcachedCache",
    "RedisCache",
    "ValkeyCache",
    "NullCache",
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

from opteryx.compiled.structures.memory_pool import MemoryPool
from opteryx.shared.async_memory_pool import AsyncMemoryPool
from opteryx.shared.buffer_pool import BufferPool
from opteryx.shared.materialized_datasets import MaterializedDatasets

__all__ = ("AsyncMemoryPool", "BufferPool", "MaterializedDatasets", "MemoryPool")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# See the License at http://www.apache.org/licenses/LICENSE-2.0
# Distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND.

"""
Optimization Rule - Limit Pushdown

Type: Heuristic
Goal: Reduce Rows

We try to push the limit to the other side of PROJECTS
"""

from typing import Optional
from typing import Set

from opteryx.connectors.capabilities import LimitPushable
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext
from .optimization_strategy import get_nodes_of_type_from_logical_plan


class LimitPushdownStrategy(OptimizationStrategy):
    """Push LIMIT operators towards scans when it is safe to do so."""

    _BARRIER_TYPES = {
        LogicalPlanStepType.Aggregate,
        LogicalPlanStepType.AggregateAndGroup,
        LogicalPlanStepType.Distinct,
        LogicalPlanStepType.Filter,
        LogicalPlanStepType.FunctionDataset,
        LogicalPlanStepType.HeapSort,
        LogicalPlanStepType.Limit,
        LogicalPlanStepType.MetadataWriter,
        LogicalPlanStepType.Order,
        LogicalPlanStepType.Set,
        LogicalPlanStepType.Union,
    }

    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore[arg-type]

        if node.node_type == LogicalPlanStepType.Limit:
            if node.offset is not None or node.limit in (None, 0):
                return context
            node.nid = context.node_id
            if not hasattr(node, "pushdown_targets"):
                node.pushdown_targets = set(node.all_relations or [])
            context.collected_limits.append(node)
            return context

        remaining_limits = []
        for limit_node in context.collected_limits:
            if self._should_skip_branch(limit_node, node):
                remaining_limits.append(limit_node)
                continue

            if node.node_type == LogicalPlanStepType.Scan:
                outcome = self._apply_to_scan(limit_node, node, context)
                if outcome is True:
                    continue
                if outcome is None:
                    remaining_limits.append(limit_node)
                    continue
                self._place_before_node(limit_node, node, context)
                continue

            if node.node_type == LogicalPlanStepType.Join:
                if self._refine_targets_for_join(limit_node, node):
                    remaining_limits.append(limit_node)
                    continue
                self._place_before_node(limit_node, node, context)
                continue

            if node.node_type in self._BARRIER_TYPES:
                self._place_before_node(limit_node, node, context)
                continue

            remaining_limits.append(limit_node)

        context.collected_limits = remaining_limits
        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        context.collected_limits.clear()
        return plan

    def should_i_run(self, plan: LogicalPlan) -> bool:
        candidates = get_nodes_of_type_from_logical_plan(plan, (LogicalPlanStepType.Limit,))
        return len(candidates) > 0

    @staticmethod
    def _collect_relations(node: LogicalPlanNode) -> Set[str]:
        relations = getattr(node, "all_relations", None)
        if relations:
            return set(relations)
        return set()

    def _should_skip_branch(self, limit_node: LogicalPlanNode, node: LogicalPlanNode) -> bool:
        targets: Set[str] = getattr(limit_node, "pushdown_targets", set())
        if not targets:
            return False
        node_relations = self._collect_relations(node)
        return bool(node_relations) and targets.isdisjoint(node_relations)

    def _apply_to_scan(
        self,
        limit_node: LogicalPlanNode,
        scan_node: LogicalPlanNode,
        context: OptimizerContext,
    ) -> Optional[bool]:
        targets: Set[str] = getattr(
            limit_node, "pushdown_targets", set(limit_node.all_relations or [])
        )
        relation_names = {scan_node.relation, getattr(scan_node, "alias", None)}
        if targets and targets.isdisjoint({name for name in relation_names if name}):
            return None

        connector = getattr(scan_node, "connector", None)
        if connector and LimitPushable in connector.__class__.mro():
            current_limit = getattr(scan_node, "limit", None)
            scan_node.limit = (
                limit_node.limit if current_limit is None else min(current_limit, limit_node.limit)
            )
            if limit_node.nid in context.optimized_plan:
                context.optimized_plan.remove_node(limit_node.nid, heal=True)
            context.optimized_plan[context.node_id] = scan_node
            self.statistics.optimization_limit_pushdown += 1
            return True

        return False

    def _refine_targets_for_join(
        self, limit_node: LogicalPlanNode, join_node: LogicalPlanNode
    ) -> bool:
        join_type = getattr(join_node, "type", None)
        if not join_type:
            return False

        targets: Set[str] = getattr(
            limit_node, "pushdown_targets", set(limit_node.all_relations or [])
        )
        if not targets:
            targets = set(limit_node.all_relations or [])

        left_relations = set(getattr(join_node, "left_relation_names", []) or [])
        right_relations = set(getattr(join_node, "right_relation_names", []) or [])

        new_targets: Optional[Set[str]] = None

        if join_type == "left outer":
            new_targets = targets & left_relations
        elif join_type == "right outer":
            new_targets = targets & right_relations
        elif join_type == "cross join":
            left_size = getattr(join_node, "left_size", float("inf"))
            right_size = getattr(join_node, "right_size", float("inf"))
            left_choice = targets & left_relations
            right_choice = targets & right_relations
            if left_choice and right_choice:
                new_targets = left_choice if left_size <= right_size else right_choice
            elif left_choice:
                new_targets = left_choice
            elif right_choice:
                new_targets = right_choice
        else:
            return False

        if not new_targets:
            return False

        limit_node.pushdown_targets = new_targets
        limit_node.all_relations = set(new_targets)
        return True

    def _place_before_node(
        self, limit_node: LogicalPlanNode, _: LogicalPlanNode, context: OptimizerContext
    ) -> None:
        if limit_node.nid in context.optimized_plan:
            context.optimized_plan.remove_node(limit_node.nid, heal=True)
        context.optimized_plan.insert_node_after(limit_node.nid, limit_node, context.node_id)
        limit_node.columns = []
        limit_node.pushdown_targets = set(limit_node.all_relations or [])
        self.statistics.optimization_limit_pushdown += 1
